import os

from kih_api import global_common

email_account = global_common.get_environment_variable("KIH_API_EMAIL_ACCOUNT")
email_account_password = global_common.get_environment_variable("KIH_API_EMAIL_ACCOUNT_PASSWORD")

execution_type__Email_Server = "Email Server"



import mongoengine
from pymongo import MongoClient

from kih_api.database import constants
from kih_api.global_common import timed

mongo_client: MongoClient = None


@timed
def connect_to_database() -> None:
    global mongo_client
    if mongo_client is None:
        mongo_client = mongoengine.connect(host=constants.DATABASE_URI, db=constants.DATABASE_NAME)



import datetime
import enum
import uuid
from dataclasses import dataclass
from decimal import Decimal
from typing import List

import pytz

import kih_api
from kih_api import global_common, communication
from kih_api.communication import telegram
from kih_api.ibkr_web_api import ibkr_models, common
from kih_api.ibkr_web_api.exceptions import StockNotFoundException, StockDataNotAvailableException
from kih_api.ibkr_web_api.ibkr_models import StockSearchResults, Contract, MarketDataSnapshot, PortfolioPositions, PortfolioAccounts, \
    MarketDataHistory, LiveOrders, LiveOrder


class InstrumentType(enum.Enum):
    STOCK: str = "STK"
    CASH: str = "CASH"
    OPTION: str = "OPT"
    INDEX: str = "IND"


class OrderType(enum.Enum):
    MARKET: str = "MKT"
    LIMIT: str = "LMT"
    STOP: str = "STP"
    STOP_LIMIT: str = "STOP_LIMIT"
    MIDPRICE: str = "MIDPRICE"


class OrderSide(enum.Enum):
    BUY: str = "BUY"
    SELL: str = "SELL"


class OrderTimeInForce(enum.Enum):
    GOOD_TILL_CANCEL: str = "GTC"
    DAY: str = "DAY"


class StockExchanges(enum.Enum):
    SMART: str = "SMART"
    NASDAQ: str = "NASDAQ"
    AMEX: str = "AMEX"
    NYSE: str = "NYSE"
    CBOE: str = "CBOE"


class OrderStatus(enum.Enum):
    SUBMITTED: str = "Submitted"
    INACTIVE: str = "Inactive"
    CANCELLED: str = "Cancelled"
    PENDING_SUBMIT: str = "PendingSubmit"
    FILLED: str = "Filled"


@dataclass
class Instrument:
    symbol: str
    contract_id: int
    instrument_type: InstrumentType
    exchanges_list: List[str]
    main_exchange: str
    name: str
    open_price: Decimal
    close_price: Decimal
    high_price: Decimal
    low_price: Decimal
    last_price: Decimal
    change_in_currency: Decimal
    change_in_percentage: Decimal
    bid_price: Decimal
    ask_price: Decimal
    ask_size: Decimal
    bid_size: Decimal
    volume: Decimal
    option_volume: Decimal
    dividend_amount: Decimal
    dividend_yield: Decimal
    market_capitalization: Decimal
    price_to_earnings_ratio: Decimal
    earnings_per_share: Decimal

    def __init__(self, symbol: str, contract_id: int, instrument_type: str, exchanges_list: List[str], main_exchange:
    str, name: str, open_price: str, close_price: str, high_price: str, low_price: str, last_price: str,
                 change_in_currency: str, change_in_percentage: str, bid_price: str, ask_price: str,
                 ask_size: str, bid_size: str,
                 volume: str, option_volume: str, dividend_amount: str, dividend_yield: str, market_capitalization:
            str, price_to_earnings_ratio: str, earnings_per_share: str):

        if contract_id is None:
            raise StockDataNotAvailableException()

        self.last_price = Decimal(str(last_price.replace("C", "").replace("H", "")))
        self.symbol = symbol

        if contract_id is not None:
            self.contract_id = contract_id

        if exchanges_list is not None:
            self.exchanges_list = exchanges_list

        if name is not None:
            self.name = name

        if instrument_type is not None:
            self.instrument_type = global_common.get_enum_from_value(instrument_type, InstrumentType)

        if main_exchange is not None:
            self.main_exchange = main_exchange

        if change_in_percentage is not None:
            self.change_in_percentage = common.extract_decimal(change_in_percentage)

        if change_in_currency is not None:
            self.change_in_currency = common.extract_decimal(change_in_currency)

        if bid_price is not None:
            self.bid_price = common.extract_decimal(bid_price)

        if bid_size is not None:
            self.bid_size = common.extract_decimal(bid_size)

        if ask_price is not None:
            self.ask_price = common.extract_decimal(ask_price)

        if ask_size is not None:
            self.ask_size = common.extract_decimal(ask_size)

        if volume is not None:
            self.volume = common.extract_decimal(volume)

        if dividend_amount is not None:
            self.dividend_amount = common.extract_decimal(dividend_amount)

        if earnings_per_share is not None:
            self.earnings_per_share = common.extract_decimal(earnings_per_share)

        if price_to_earnings_ratio is not None:
            self.price_to_earnings_ratio = common.extract_decimal(price_to_earnings_ratio)

        if market_capitalization is not None:
            self.market_capitalization = common.extract_decimal(market_capitalization)

        if open_price is not None:
            self.open_price = common.extract_decimal(open_price)

        if low_price is not None:
            self.low_price = common.extract_decimal(low_price)

        if high_price is not None:
            self.high_price = common.extract_decimal(high_price)

        if close_price is not None:
            self.close_price = common.extract_decimal(close_price)

        if option_volume is not None:
            self.option_volume = common.extract_decimal(option_volume)

        if dividend_yield is not None:
            self.dividend_yield = common.extract_decimal(str(dividend_yield).replace("%", ""))

    @classmethod
    def get(cls, symbol: str, instrument_type: InstrumentType, exchange: StockExchanges, number_of_tries: int = 0) -> "Instrument":
        stock_search_results: StockSearchResults = StockSearchResults.call(symbol)
        contract_id: int = None

        if stock_search_results.stock_search_results_list is None:
            raise StockNotFoundException()

        for stock_search_result in stock_search_results.stock_search_results_list:
            if stock_search_result.description == exchange.value:
                contract_id = stock_search_result.conid
                break

        contract: Contract = Contract.call(contract_id)
        market_data_snapshot: MarketDataSnapshot = MarketDataSnapshot.call(contract_id)

        if market_data_snapshot.last_price is None:
            if number_of_tries <= kih_api.ibkr_web_api.constants.NUMBER_OF_RE_TRIES:
                return Instrument.get(symbol, instrument_type, exchange, number_of_tries + 1)
            else:
                raise StockDataNotAvailableException()

        if contract.instrument_type != instrument_type.value:
            raise StockNotFoundException()

        return Instrument(contract.symbol, contract.con_id, contract.instrument_type, contract.valid_exchanges,
                          contract.exchange, contract.company_name, market_data_snapshot.open_price,
                          market_data_snapshot.close_price, market_data_snapshot.current_day_high_price,
                          market_data_snapshot.current_day_low_price, market_data_snapshot.last_price,
                          market_data_snapshot.change_in_currency, market_data_snapshot.change_in_percentage,
                          market_data_snapshot.bid_price, market_data_snapshot.ask_price, market_data_snapshot.ask_size,
                          market_data_snapshot.bid_size, market_data_snapshot.volume, market_data_snapshot.option_volume,
                          market_data_snapshot.dividend_amount, market_data_snapshot.dividend_yield,
                          market_data_snapshot.market_capitalization, market_data_snapshot.price_to_earnings_ratio,
                          market_data_snapshot.earnings_per_share)


@dataclass
class HistoricalData:
    symbol: str
    high: Decimal
    low: Decimal
    open: Decimal
    close: Decimal
    timestamp: datetime.datetime

    def __init__(self, ticker: str, high: Decimal, low: Decimal, open_price: Decimal, close: Decimal, timestamp: datetime.datetime):
        self.ticker = ticker
        self.high = high
        self.low = low
        self.open = open_price
        self.close = close
        self.timestamp = timestamp

    @classmethod
    def get(cls, symbol: str, instrument_type: InstrumentType, exchange: StockExchanges, period: str = "10y", bar: str = "1d") -> List["HistoricalData"]:
        stock: Instrument = Instrument.get(symbol, instrument_type, exchange)
        market_data_history: MarketDataHistory = MarketDataHistory.call(stock.contract_id, period, bar)
        stock_history_list: List[HistoricalData] = []
        for past_market_data in market_data_history.market_data_history:
            stock_history_list.append(HistoricalData(symbol, past_market_data.high, past_market_data.low,
                                                     past_market_data.open, past_market_data.close,
                                                     past_market_data.timestamp))
        return stock_history_list


@dataclass
class Position:
    description: str
    position_size: Decimal
    market_price: Decimal
    market_value: Decimal
    instrument_type: InstrumentType
    currency: global_common.Currency
    account_id: str

    def __init__(self, description: str, position_size: float, market_price: float, market_value: float,
                 instrument_type: str, currency: str, account_id: str):
        self.description = description
        self.position_size = Decimal(str(position_size))
        self.market_price = Decimal(str(market_price))
        self.market_value = Decimal(str(market_value))
        self.account_id = account_id
        self.instrument_type = global_common.get_enum_from_value(instrument_type, InstrumentType)
        self.currency = global_common.get_enum_from_value(currency, global_common.Currency)

    @classmethod
    def get_all_by_account_id(cls, account_id: str) -> List["Position"]:
        portfolio_positions: PortfolioPositions = PortfolioPositions.call(account_id)
        all_positions: List[Position] = []

        if portfolio_positions.portfolio_positions_list is not None:
            for portfolio_position in portfolio_positions.portfolio_positions_list:
                all_positions.append(Position(portfolio_position.contractDesc,
                                              portfolio_position.position, portfolio_position.mktPrice,
                                              portfolio_position.mktValue, portfolio_position.assetClass,
                                              portfolio_position.currency, account_id))
        return all_positions


@dataclass
class Account:
    account_id: str
    alias: str

    def __init__(self, account_id: str, alias: str):
        self.account_id = account_id
        self.alias = alias

    @classmethod
    def get_all(cls) -> List["Account"]:
        portfolio_accounts: PortfolioAccounts = PortfolioAccounts.call()
        all_accounts_list: List[Account] = []
        for portfolio_account in portfolio_accounts.portfolio_account_list:
            all_accounts_list.append(Account(portfolio_account.accountId, portfolio_account.accountAlias))

        return all_accounts_list


@dataclass
class PlaceOrder:
    symbol: str
    secType: InstrumentType
    orderType: OrderType
    side: OrderSide
    quantity: Decimal
    account_id: str
    custom_order_id: str
    price: Decimal = None
    outsideRTH: bool = True
    tif: OrderTimeInForce = OrderTimeInForce.GOOD_TILL_CANCEL

    def __init__(self, symbol: str, orderType: OrderType, side: OrderSide, quantity: int, price: float, account_id: str):

        self.symbol = symbol
        self.orderType = orderType
        self.side = side
        self.quantity = Decimal(str(quantity))

        if orderType != OrderType.MARKET:
            self.price = Decimal(str(price))
        else:
            self.outsideRTH = False

        self.account_id = account_id
        self.custom_order_id = str(uuid.uuid4())

    def execute(self) -> "PlaceOrderResponse":
        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Placing a new order</b></u>\n\nSymbol: <i>{self.symbol}</i>"
                                            f"\nOrder Type: <i>{self.orderType.value}</i>"
                                            f"\nOrder Side: <i>{self.side.value}</i>"
                                            f"\nQuantity: <i>{str(self.quantity)}</i>\nPrice: "
                                            f"<i>{str(self.price)}</i>\nAccount ID: <i>{self.account_id}</i>", True)

        stock: Instrument = Instrument.get(self.symbol, InstrumentType.STOCK, StockExchanges.NASDAQ)

        price: float = None
        if self.price is not None:
            price = float(self.price)

        order_response: PlaceOrderResponse = PlaceOrderResponse(ibkr_models.PlaceOrder.call(self.account_id,
                                                                                            stock.contract_id,
                                                                                            self.orderType.value, self.outsideRTH,
                                                                                            self.side.value, self.tif.value,
                                                                                            int(self.quantity), self.custom_order_id,
                                                                                            price))

        if order_response.is_order_filled:
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>Order has been placed and filled</b></u>"
                                                f"\n\nSymbol: <i>{self.symbol}</i>"
                                                f"\nOrder Type: <i>{self.orderType.value}</i>"
                                                f"\nQuantity: <i>{str(self.quantity)}</i>"
                                                f"\nPrice: <i>{str(price)}</i>"
                                                f"\nAccount ID: <i>{self.account_id}</i>", True)
        elif order_response.is_order_placed:
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>Order has been placed (but not filled)</b></u>"
                                                f"\n\nSymbol: <i>{self.symbol}</i>"
                                                f"\nOrder Type: <i>{self.orderType.value}</i>"
                                                f"\nQuantity: <i>{str(self.quantity)}</i>"
                                                f"\nPrice: <i>{str(price)}</i>"
                                                f"\nAccount ID: <i>{self.account_id}</i>", True)
        else:
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Order has not been placed</b></u>"
                                                f"\n\nSymbol: <i>{self.symbol}</i>"
                                                f"\nOrder Type: <i>{self.orderType.value}</i>"
                                                f"\nQuantity: <i>{str(self.quantity)}</i>"
                                                f"\nPrice: <i>{str(price)}</i>"
                                                f"\nAccount ID: <i>{self.account_id}</i>"
                                                f"\nResponse Message: <i>{kih_api.ibkr_web_api.common.get_html_commented(order_response.response_text)}</i>",
                                                True)

        return order_response


@dataclass
class CancelOrder:
    unfilled_order: "UnfilledOrder"
    account_id: str
    order_id: str

    def __init__(self, unfilled_order: "UnfilledOrder"):
        self.unfilled_order = unfilled_order
        self.account_id = self.unfilled_order.account_id
        self.order_id = self.unfilled_order.order_id

    def execute(self) -> "CancelOrderResponse":
        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Cancelling an unfilled order</b></u>"
                                            f"\n\nSymbol: <i>{self.unfilled_order.symbol}</i>\n"
                                            f"\nUnfilled Quantity: <i>{str(self.unfilled_order.unfilled_quantity)}</i>"
                                            f"\nOrder Type: {self.unfilled_order.order_type.value}"
                                            f"\nPrice: <i>{str(self.unfilled_order.price)}</i>"
                                            f"\nAccount ID: <i>{self.account_id}</i>", True)

        cancel_order_response: CancelOrderResponse = CancelOrderResponse(
            kih_api.ibkr_web_api.ibkr_models.CancelOrder.call(self.account_id, self.order_id))

        if not cancel_order_response.is_order_cancelled:
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Order has not been cancelled</b></u>"
                                                f"\n\nSymbol: <i>{self.unfilled_order.symbol}</i>\n"
                                                f"\nUnfilled Quantity: <i>{str(self.unfilled_order.unfilled_quantity)}</i>"
                                                f"\nOrder Type: {self.unfilled_order.order_type.value}"
                                                f"\nPrice: <i>{str(self.unfilled_order.price)}</i>"
                                                f"\nAccount ID: <i>{self.account_id}</i>", True)

        return cancel_order_response

    @classmethod
    def all_unfilled_orders(cls) -> None:
        for unfilled_order in UnfilledOrder.get():
            cancel_order: CancelOrderResponse = CancelOrder(unfilled_order).execute()


class CancelOrderResponse:
    is_order_cancelled: bool
    order_id: str
    contract_id: int
    account_id: str
    response_message: str

    def __init__(self, cancel_order_response: kih_api.ibkr_web_api.ibkr_models.CancelOrderResponse):
        self.order_id = cancel_order_response.order_id
        self.contract_id = cancel_order_response.conid
        self.account_id = cancel_order_response.account
        self.response_message = cancel_order_response.msg
        self.is_order_cancelled = cancel_order_response.is_successful


@dataclass
class PlaceOrderResponse:
    is_order_placed: bool
    ibkr_order_id: str
    order_id: str
    order_cancelled: datetime.datetime
    response_text: str
    is_order_filled: bool

    def __init__(self, order_response: ibkr_models.PlaceOrderResponse):
        self.is_order_placed = order_response.is_successful and (order_response.order_status == OrderStatus.SUBMITTED.value or order_response.order_status == OrderStatus.FILLED.value)
        self.ibkr_order_id = order_response.order_id
        self.order_id = order_response.order_id
        self.response_text = order_response.text

        if order_response.text is not None and "will be automatically canceled at 20220101 13:00:00 HKT" in \
                order_response.text:
            datetime_string: str = order_response.text.split("will be automatically canceled at ")[-1].replace(" HKT",
                                                                                                               "")
            self.order_cancelled = pytz.timezone("Asia/Hong_Kong").localize(
                datetime.datetime.strptime(datetime_string, "%Y%m%d %H:%M:%S"))

        for live_order in LiveOrders.call().live_orders_list:
            if str(live_order.orderId) == order_response.order_id:
                self.is_order_filled = True
                break
            else:
                self.is_order_filled = False


@dataclass
class UnfilledOrder:
    account_id: str
    exchange: StockExchanges
    contract_id: int
    order_id: str
    currency: global_common.Currency
    unfilled_quantity: Decimal
    filled_quantity: Decimal
    order_description: str
    symbol: str
    security_type: InstrumentType
    status: str
    order_type: OrderType
    last_execution_time: datetime.datetime
    custom_order_id: str
    price: Decimal
    time_in_force: OrderTimeInForce
    side: OrderSide

    def __init__(self, live_order: LiveOrder):
        self.account_id = live_order.acct
        self.exchange = global_common.get_enum_from_value(live_order.exchange, StockExchanges)
        self.contract_id = live_order.conid
        self.order_id = live_order.orderId
        self.currency = global_common.get_enum_from_value(live_order.cashCcy, global_common.Currency)
        self.unfilled_quantity = Decimal(str(live_order.remainingQuantity))
        self.filled_quantity = Decimal(str(live_order.filledQuantity))
        self.order_description = live_order.orderDesc
        self.symbol = live_order.ticker
        self.security_type = global_common.get_enum_from_value(live_order.secType, InstrumentType)
        self.status = live_order.status
        self.last_execution_time = datetime.datetime.utcfromtimestamp(int(live_order.lastExecutionTime_r / 1000))
        self.custom_order_id = live_order.order_ref
        self.time_in_force = global_common.get_enum_from_value(live_order.timeInForce, OrderTimeInForce)
        self.side = global_common.get_enum_from_value(live_order.side, OrderSide)

        if live_order.orderType == "Limit":
            self.order_type = OrderType.LIMIT
        elif live_order.orderType == "Market":
            self.order_type = OrderType.MARKET

        if self.order_type is not OrderType.MARKET:
            self.price = Decimal(str(live_order.price))
        else:
            self.price = None

    @classmethod
    def get(cls) -> List["UnfilledOrder"]:
        live_orders: LiveOrders = LiveOrders.call()

        unfilled_orders_list: List[UnfilledOrder] = []
        for live_order in live_orders.live_orders_list:
            if live_order.status != OrderStatus.INACTIVE.value and live_order.status != OrderStatus.CANCELLED.value and live_order.status != OrderStatus.FILLED.value:
                unfilled_orders_list.append(UnfilledOrder(live_order))

        return unfilled_orders_list

    @classmethod
    def get_all_unfilled_orders_value(cls, exchange: StockExchanges) -> Decimal:
        value: Decimal = Decimal("0")
        for unfilled_order in UnfilledOrder.get():
            value_of_order: Decimal = None
            if unfilled_order.order_type == OrderType.LIMIT:
                value_of_order = unfilled_order.price * unfilled_order.unfilled_quantity
            elif unfilled_order.order_type == OrderType.MARKET:
                value_of_order = Instrument.get(unfilled_order.symbol, InstrumentType.STOCK, exchange).last_price * \
                                 unfilled_order.unfilled_quantity

            value = value + value_of_order

        return value


@dataclass
class AccountInformation:
    available_funds: Decimal
    currency: global_common.Currency
    updated_at: datetime.datetime
    net_liquidity: Decimal

    def __init__(self, account_information: ibkr_models.AccountInformation):
        self.available_funds = Decimal(str(account_information.totalcashvalue.amount))
        self.currency = global_common.get_enum_from_value(account_information.fullavailablefunds.currency, global_common.Currency)
        self.updated_at = datetime.datetime.utcfromtimestamp(int(account_information.fullavailablefunds.timestamp / 1000))
        self.net_liquidity = Decimal(str(account_information.netliquidation.amount))

    @classmethod
    def get_by_account_id(cls, account_id: str) -> "AccountInformation":
        return AccountInformation(ibkr_models.AccountInformation.call(account_id))


@dataclass
class PortfolioPosition:
    name: str
    full_name: str
    symbol: str
    account_id: str
    contract_id: int
    position: Decimal
    market_price: Decimal
    market_value: Decimal
    currency: global_common.Currency
    instrument_type: InstrumentType

    def __init__(self, portfolio_position: ibkr_models.PortfolioPosition):
        self.name = portfolio_position.name
        self.full_name = portfolio_position.fullName
        self.account_id = portfolio_position.acctId
        self.symbol = portfolio_position.contractDesc
        self.contract_id = portfolio_position.conid
        self.market_price = Decimal(str(portfolio_position.mktPrice))
        self.market_value = Decimal(str(portfolio_position.mktValue))
        self.currency = global_common.get_enum_from_value(portfolio_position.currency, global_common.Currency)
        self.instrument_type = global_common.get_enum_from_value(portfolio_position.assetClass, InstrumentType)
        self.position = Decimal(str(portfolio_position.position))

    @classmethod
    def get_by_account_id(cls, account_id: str) -> List["PortfolioPosition"]:
        portfolio_positions: ibkr_models.PortfolioPositions = ibkr_models.PortfolioPositions.call(account_id)

        portfolio_position_list: List[PortfolioPosition] = []
        for portfolio_position in portfolio_positions.portfolio_positions_list:
            portfolio_position_list.append(PortfolioPosition(portfolio_position))

        return portfolio_position_list

    @classmethod
    def close_all_by_account_id(cls, account_id: str) -> List[PlaceOrderResponse]:
        all_positions: List["PortfolioPosition"] = PortfolioPosition.get_by_account_id(account_id)
        place_order_response_list: List[PlaceOrderResponse] = []
        is_all_orders_placed: bool = True

        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Closing all positions</b></u>"
                                            f"\n\nAccount ID: <i>{account_id}</i>\n", True)

        for position in all_positions:
            if position.position == 0:
                continue

            order_side: OrderSide = OrderSide.SELL if position.position > 0 else OrderSide.BUY

            place_order_response: PlaceOrderResponse = PlaceOrder(position.symbol, OrderType.MARKET, order_side, abs(int(position.position)), None, account_id).execute()
            place_order_response_list.append(place_order_response)
            if not place_order_response.is_order_placed:
                is_all_orders_placed = False

        if not is_all_orders_placed:
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: All positions failed to close</b></u>"
                                                f"\n\nAccount ID: <i>{account_id}</i>\n", True)
        else:
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<i>Orders placed to close all positions</i>"
                                                f"\nAccount ID: <i>{account_id}</i>\n", True)

        return place_order_response_list



KEEP_ALIVE_TIMEOUT_IN_SECONDS: int = 60
ENDPOINT_BASE: str = "https://localhost:5000/v1/api/"
ENDPOINT_KEEP_ALIVE: str = ENDPOINT_BASE + "tickle"
ENDPOINT_AUTHENTICATION_STATUS: str = ENDPOINT_BASE + "iserver/auth/status"
ENDPOINT_RE_AUTHENTICATE: str = ENDPOINT_BASE + "iserver/reauthenticate?force=true"
ENDPOINT_PORTFOLIO_ACCOUNTS: str = ENDPOINT_BASE + "portfolio/accounts"
ENDPOINT_PORTFOLIO_POSITIONS: str = ENDPOINT_BASE + "portfolio/{accountId}/positions/0"
ENDPOINT_STOCK_SEARCH: str = ENDPOINT_BASE + "iserver/secdef/search"
ENDPOINT_CONTRACT: str = ENDPOINT_BASE + "iserver/contract/{contract_id}/info"
ENDPOINT_MARKET_DATA_HISTORY: str = ENDPOINT_BASE + "iserver/marketdata/history?conid={contract_id}&period={" \
                                                    "period}&bar={bar}"
ENDPOINT_MARKET_DATA_SNAPSHOT: str = ENDPOINT_BASE + "/iserver/marketdata/snapshot?conids={contract_id}&fields=31," \
                                                     "55,58,70,71,72,73,74,75,76,77,78,82,83,84,85,86,87,88,6004,6008,6070,6072,6073,6119,6457,6509,7051,7057,7058,7059,7068,7084,7085,7086,7087,7088,7089,7094,7219,7220,7221,7280,7281,7282,7283,7284,7285,7286,7287,7288,7289,7290,7291,7292,7293,7294,7295,7296,7308,7309,7310,7311,7607,7633,7635,7636,7637,7638,7639,7644,7655,7671,7672,7674,7675,7676,7677,7678,7679,7680,7681,7682,7683,7684,7685,7686,7687,7688,7689,7690,7694,7695,7696,7697,7698,7699,7700,7702,7703,7704,7705,7706,7707,7708,7714,7715,7718,7720,7741,7762"
ENDPOINT_LIST_OF_TRADES: str = ENDPOINT_BASE + "portal/iserver/account/trades"
ENDPOINT_CREATE_ORDER: str = ENDPOINT_BASE + "iserver/account/{account_id}/orders"
ENDPOINT_CANCEL_ORDER: str = ENDPOINT_BASE + "iserver/account/{account_id}/order/{order_id}"
ENDPOINT_ACCOUNTS: str = ENDPOINT_BASE + "iserver/accounts"
ENDPOINT_LIVE_ORDERS: str = ENDPOINT_BASE + "iserver/account/orders"
ENDPOINT_ACCOUNT_INFORMATION: str = ENDPOINT_BASE + "portfolio/{account_id}/summary"
ENDPOINT_SELECT_ACCOUNT: str = ENDPOINT_BASE + "iserver/accounts"
ENDPOINT_PLACE_ORDER_REPLY: str = ENDPOINT_BASE + "iserver/reply/{reply_id}"
NUMBER_OF_RE_TRIES: int = 3



from typing import Callable, Any

import kih_api
from kih_api import global_common, communication
from kih_api.communication import telegram
from kih_api.ibkr.models import IBKR


def job(job_name: str, is_run_only_when_market_is_open: bool = False, symbol_to_check_market_open: str = None) -> Callable:
    def decorator(func: Callable) -> Callable:
        @global_common.job(job_name)
        def wrapper(*args: Any, **kwargs: Any) -> None:
            if not is_run_only_when_market_is_open or (is_run_only_when_market_is_open and IBKR.is_market_open(symbol_to_check_market_open)):
                func(*args, **kwargs)
            else:
                telegram.send_message(
                    telegram.constants.telegram_channel_username, f"Market is not open; job skipped: <i>{job_name}</i>", True)

        return wrapper

    return decorator



import time
from typing import Dict, Optional, Any

import requests
import urllib3
from requests import Response

from kih_api.global_common import CustomException
from kih_api.http_requests import constants
from kih_api.http_requests.models import MethodType
from kih_api.logger import logger

is_ssl_certificate_verification_used: bool = True
urllib3.disable_warnings()


class ServerErrorException(CustomException):
    pass


class ClientErrorException(CustomException):
    pass


def get(url: str, parameters: Optional[Dict[str, Any]] = None, headers: Dict[str, str] = None) -> Response:
    return request(url, parameters, MethodType.GET, headers=headers)


def post(url: str, parameters: Optional[Dict[str, Any]] = None, headers: Dict[str, str] = None) -> Response:
    return request(url, parameters, MethodType.POST, headers=headers)


def put(url: str, parameters: Optional[Dict[str, Any]] = None, headers: Dict[str, str] = None) -> Response:
    return request(url, parameters, MethodType.PUT, headers=headers)


def delete(url: str, parameters: Optional[Dict[str, Any]] = None, headers: Dict[str, str] = None) -> Response:
    return request(url, parameters, MethodType.DELETE, headers=headers)


def request(url: str, parameters: Optional[Dict[str, Any]], method_type: MethodType, server_error_retry_times: int = 0, headers: Dict[str, str] = None) -> Response:
    global is_ssl_certificate_verification_used
    start_time: float = time.time()

    response: Response = None
    if method_type == MethodType.GET:
        response = requests.get(url=url, params=parameters, verify=is_ssl_certificate_verification_used, headers=headers)
    elif method_type == MethodType.POST:
        response = requests.post(url=url, json=parameters, verify=is_ssl_certificate_verification_used, headers=headers)
    elif method_type == MethodType.PUT:
        response = requests.put(url=url, params=parameters, verify=is_ssl_certificate_verification_used, headers=headers)
    elif method_type == MethodType.DELETE:
        response = requests.delete(url=url, params=parameters, verify=is_ssl_certificate_verification_used, headers=headers)

    if 400 <= response.status_code <= 499:
        logger.error("Response code: " + str(response.status_code))
        raise ClientErrorException(response.text)
    elif 500 <= response.status_code <= 599:
        logger.error("Response code: " + str(response.status_code))
        if server_error_retry_times < constants.SERVER_ERROR_RETRY_TIMES:
            logger.info("Retrying")
            response = request(url, parameters, method_type, server_error_retry_times + 1)
        else:
            raise ServerErrorException(response.text)

    logger.debug(
        constants.EXECUTION_TYPE_API_Call + " | " + method_type.value + " | " + response.url + " | " + str(time.time() - start_time) + " | " + url + " | " + str(response.status_code))
    logger.debug(response.text)

    return response



import datetime
import os
import shutil
import tempfile

import requests
import validators
from dateutil.relativedelta import relativedelta

from kih_api.finance_database.exceptions import ExcelFileOpenedByAnotherApplication


def get_temp_file_path(file_path: str) -> str:
    temp_dir = tempfile.gettempdir()
    temp_file_path = os.path.join(temp_dir, 'temp_file_name')
    try:
        if validators.url(file_path):
            open(temp_file_path, 'wb').write(requests.get(file_path).content)
        else:
            shutil.copy(file_path, temp_file_path)
    except PermissionError:
        raise ExcelFileOpenedByAnotherApplication()
    return temp_file_path


def get_next_month_text() -> str:
    today: datetime.date = datetime.date.today()
    start_of_this_month: datetime.date = datetime.date(today.year, today.month, 1)
    start_of_next_month: datetime.date = start_of_this_month + relativedelta(months=1)
    return start_of_next_month.strftime("%B, %Y")



import os

from kih_api import global_common

IP_ADDRESS_INFO_ENDPOINT: str = "http://ip-api.com/json/"
ENVIRONMENT: str = "DEV" if global_common.get_environment_variable("ENV") is None else global_common.get_environment_variable("ENV")



import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, List, Dict, Any

from requests import Response

from kih_api import global_common, http_requests
from kih_api.http_requests import common
from kih_api.http_requests.models import ResponseObject
from kih_api.wise import constants


def get_header(api_key: str) -> Dict[str, str]:
    return {"Authorization": f"Bearer {api_key}"}

@dataclass
class UserProfileDetails:
    occupations: None = None
    acn: None = None
    abn: None = None
    arbn: None = None
    webpage: None = None
    businessSubCategory: Optional[str] = None
    avatar: Optional[str] = None
    occupation: Optional[str] = None
    firstName: Optional[str] = None
    lastName: Optional[str] = None
    dateOfBirth: Optional[str] = None
    phoneNumber: Optional[str] = None
    primaryAddress: Optional[int] = None
    name: Optional[str] = None
    registrationNumber: Optional[str] = None
    companyType: Optional[str] = None
    companyRole: Optional[str] = None
    descriptionOfBusiness: Optional[str] = None
    businessCategory: Optional[str] = None
    firstNameInKana: Optional[str] = None
    lastNameInKana: Optional[str] = None


@dataclass
class UserProfiles(ResponseObject):
    id: Optional[int] = None
    type: Optional[str] = None
    details: Optional[UserProfileDetails] = None
    endpoint: str = constants.ENDPOINT_PROFILES

    @classmethod
    def call(cls, api_key: str) -> List["UserProfiles"]:
        response: Response = http_requests.get(cls.endpoint, headers=get_header(api_key))
        return common.get_model_from_response(response, cls)  # type: ignore


@dataclass
class Amount:
    value: Optional[float] = None
    currency: Optional[str] = None


@dataclass
class BankAddress:
    stateCode: None = None
    postCode: Optional[int] = None
    addressFirstLine: Optional[str] = None
    city: Optional[str] = None
    country: Optional[str] = None


@dataclass
class BankDetails:
    id: Optional[int] = None
    currency: Optional[str] = None
    bankCode: Optional[str] = None
    accountNumber: Optional[str] = None
    swift: Optional[str] = None
    iban: Optional[str] = None
    bankName: Optional[str] = None
    accountHolderName: Optional[str] = None
    bankAddress: Optional[BankAddress] = None


@dataclass
class Balance:
    name: Optional[str] = None
    icon: Optional[Dict] = None
    id: Optional[int] = None
    currency: Optional[str] = None
    amount: Optional[Amount] = None
    reservedAmount: Optional[Amount] = None
    cashAmount: Optional[Amount] = None
    totalWorth: Optional[Amount] = None
    type: Optional[str] = None
    investmentState: Optional[str] = None
    creationTime: Optional[str] = None
    modificationTime: Optional[str] = None
    visible: Optional[bool] = None
    primary: Optional[bool] = None

    def __init__(self, id: int, balanceType: str, currency: str, amount: Dict, reservedAmount: Dict[str, Any], bankDetails: Dict[str, Any]):
        self.id = id
        self.balanceType = balanceType
        self.currency = currency
        self.amount = Amount(**amount)
        self.reservedAmount = Amount(**reservedAmount)


@dataclass
class Account(ResponseObject):
    name: Optional[str] = None
    icon: Optional[Dict] = None
    id: Optional[int] = None
    currency: Optional[str] = None
    amount: Optional[Amount] = None
    reservedAmount: Optional[Amount] = None
    cashAmount: Optional[Amount] = None
    totalWorth: Optional[Amount] = None
    type: Optional[str] = None
    investmentState: Optional[str] = None
    creationTime: Optional[str] = None
    modificationTime: Optional[str] = None
    visible: Optional[bool] = None
    primary: Optional[bool] = None
    endpoint: str = constants.ENDPOINT_ACCOUNTS
    endpoint_create_account: str = constants.ENDPOINT_CREATE_ACCOUNT

    @classmethod
    def call(cls, api_key: str, profile_id: int) -> List["Account"]:
        from kih_api.wise.models import AccountType

        response: Response = http_requests.get(cls.endpoint.replace("{profile_id}", str(profile_id)).replace("{account_type}", AccountType.CashAccount.value), headers=get_header(api_key))
        all_accounts_list: List[Account] = common.get_model_from_response(response, cls)  # type: ignore

        response = http_requests.get(cls.endpoint.replace("{profile_id}", str(profile_id)).replace("{account_type}", AccountType.ReserveAccount.value), headers=get_header(api_key))
        all_accounts_list.extend(common.get_model_from_response(response, cls))  # type: ignore

        return all_accounts_list

    @classmethod
    def call_create_reserve_account(cls, api_key: str, name: str, currency: global_common.Currency, profile_id: int) -> None:
        from kih_api.wise.models import AccountType

        parameters: Dict[str, Any] = {
            "currency": currency.value,
            "type": AccountType.ReserveAccount.value,
            "name": name
        }

        headers = get_header(api_key)
        headers["X-idempotence-uuid"] = str(uuid.uuid4())
        http_requests.post(cls.endpoint_create_account.replace("{profile_id}", str(profile_id)), parameters=parameters, headers=headers)


@dataclass
class ExchangeRate(ResponseObject):
    rate: Optional[float] = None
    source: Optional[str] = None
    target: Optional[str] = None
    time: Optional[str] = None
    endpoint: str = constants.ENDPOINT_EXCHANGE_RATES

    @classmethod
    def call(cls, api_key: str, source_currency: str = None, target_currency: str = None) -> "ExchangeRate":

        if source_currency is not None and target_currency is not None:
            response: Response = http_requests.get(cls.endpoint + f"?source={source_currency}&target={target_currency}", headers=get_header(api_key))
        else:
            response = http_requests.get(cls.endpoint, headers=get_header(api_key))

        return common.get_model_from_response(response, cls)[0]  # type: ignore


@dataclass
class Transfer(ResponseObject):
    transferRequest: None = None
    sourceAccount: None = None
    quote: None = None
    quoteUuid: Optional[str] = None
    id: Optional[int] = None
    user: Optional[int] = None
    targetAccount: Optional[int] = None
    status: Optional[str] = None
    reference: Optional[str] = None
    rate: Optional[float] = None
    created: Optional[str] = None
    business: Optional[int] = None
    details: Optional["Details"] = None
    hasActiveIssues: Optional[bool] = None
    sourceCurrency: Optional[str] = None
    sourceValue: Optional[float] = None
    targetCurrency: Optional[str] = None
    targetValue: Optional[float] = None
    customerTransactionId: Optional[str] = None
    endpoint: str = constants.ENDPOINT_TRANSFER

    @classmethod
    def call(cls, api_key: str, target_account_id: int, quote_uuid: str, reference: str) -> "Transfer":
        parameters: Dict[str, Any] = {
            "targetAccount": target_account_id,
            "quoteUuid": quote_uuid,
            "customerTransactionId": str(uuid.uuid4()),
            "details": {"reference": reference}
        }

        response: Response = http_requests.post(cls.endpoint, parameters=parameters, headers=get_header(api_key))
        return common.get_model_from_response(response, cls)  # type: ignore


@dataclass
class DisabledReason:
    code: Optional[str] = None
    message: Optional[str] = None
    arguments: Optional[List[Any]] = None


@dataclass
class Fee:
    transferwise: Optional[float] = None
    payIn: Optional[float] = None
    discount: Optional[int] = None
    total: Optional[float] = None
    priceSetId: Optional[int] = None
    partner: Optional[float] = None


@dataclass
class PaymentOption:
    formattedEstimatedDelivery: Optional[str] = None
    estimatedDeliveryDelays: Optional[List[Any]] = None
    allowedProfileTypes: Optional[List[str]] = None
    feePercentage: Optional[float] = None
    estimatedDelivery: Optional[str] = None
    sourceAmount: Optional[float] = None
    targetAmount: Optional[float] = None
    sourceCurrency: Optional[str] = None
    targetCurrency: Optional[str] = None
    payOut: Optional[str] = None
    fee: Optional[Fee] = None
    payIn: Optional[str] = None
    disabled: Optional[bool] = None
    disabledReason: Optional[DisabledReason] = None


@dataclass
class HighAmount:
    showFeePercentage: Optional[bool] = None
    trackAsHighAmountSender: Optional[bool] = None
    showEducationStep: Optional[bool] = None
    offerPrefundingOption: Optional[bool] = None


@dataclass
class TransferFlowConfig:
    highAmount: Optional[HighAmount] = None


@dataclass
class Quote(ResponseObject):
    targetAmount: Optional[float] = None
    guaranteedTargetAmountAllowed: Optional[bool] = None
    targetAmountAllowed: Optional[bool] = None
    paymentOptions: Optional[List[PaymentOption]] = None
    notices: Optional[List[Any]] = None
    transferFlowConfig: Optional[TransferFlowConfig] = None
    rateTimestamp: Optional[str] = None
    clientId: Optional[str] = None
    id: Optional[str] = None
    type: Optional[str] = None
    status: Optional[str] = None
    profile: Optional[int] = None
    sourceCurrency: Optional[str] = None
    targetCurrency: Optional[str] = None
    rate: Optional[float] = None
    createdTime: Optional[str] = None
    user: Optional[int] = None
    rateType: Optional[str] = None
    rateExpirationTime: Optional[str] = None
    payOut: Optional[str] = None
    guaranteedTargetAmount: Optional[bool] = None
    providedAmountType: Optional[str] = None
    expirationTime: Optional[str] = None
    targetAccount: Optional[int] = None
    payInCountry: Optional[str] = None
    payOutCountry: Optional[str] = None
    funding: Optional[str] = None
    endpoint: str = constants.ENDPOINT_QUOTE

    @classmethod
    def call(cls, api_key: str, profile_id: int, source_currency: str, target_currency: str, target_amount: float) -> "Quote":
        parameters: Dict[str, Any] = {
            "profile": profile_id,
            "sourceCurrency": source_currency,
            "targetCurrency": target_currency,
            "targetAmount": target_amount,
            "payOut": "BALANCE"
        }
        response: Response = http_requests.post(cls.endpoint.replace("{profile_id}", str(profile_id)), parameters=parameters, headers=get_header(api_key))
        return common.get_model_from_response(response, cls)  # type: ignore


@dataclass
class Address:
    country: Optional[str] = None
    countryCode: Optional[str] = None
    firstLine: Optional[str] = None
    postCode: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None


@dataclass
class RecipientDetails:
    accountHolderName: Optional[str] = None
    sortCode: Optional[str] = None
    abartn: Optional[str] = None
    accountType: Optional[str] = None
    bankgiroNumber: Optional[str] = None
    ifscCode: Optional[str] = None
    bsbCode: Optional[str] = None
    institutionNumber: Optional[str] = None
    transitNumber: Optional[str] = None
    phoneNumber: Optional[str] = None
    bankCode: Optional[str] = None
    russiaRegion: Optional[str] = None
    routingNumber: Optional[str] = None
    branchCode: Optional[str] = None
    cpf: Optional[str] = None
    cardToken: Optional[str] = None
    idType: Optional[str] = None
    idNumber: Optional[str] = None
    idCountryIso3: Optional[str] = None
    idValidFrom: Optional[str] = None
    idValidTo: Optional[str] = None
    clabe: Optional[str] = None
    dateOfBirth: Optional[str] = None
    clearingNumber: Optional[str] = None
    bankName: Optional[str] = None
    branchName: Optional[str] = None
    businessNumber: Optional[str] = None
    province: Optional[str] = None
    city: Optional[str] = None
    rut: Optional[str] = None
    token: Optional[str] = None
    cnpj: Optional[str] = None
    payinReference: Optional[str] = None
    pspReference: Optional[str] = None
    orderId: Optional[str] = None
    idDocumentType: Optional[str] = None
    idDocumentNumber: Optional[str] = None
    targetProfile: Optional[str] = None
    targetUserId: Optional[str] = None
    taxId: Optional[str] = None
    job: Optional[str] = None
    nationality: Optional[str] = None
    interacAccount: Optional[str] = None
    bban: Optional[str] = None
    town: Optional[str] = None
    postCode: Optional[str] = None
    language: Optional[str] = None
    billerCode: Optional[str] = None
    customerReferenceNumber: Optional[str] = None
    IBAN: Optional[str] = None
    iban: Optional[str] = None
    address: Optional[Address] = None
    email: Optional[str] = None
    legalType: Optional[str] = None
    accountNumber: Optional[str] = None
    swiftCode: Optional[str] = None
    bic: Optional[str] = None
    BIC: Optional[str] = None
    prefix: Optional[str] = None


@dataclass
class Recipient(ResponseObject):
    business: None = None
    id: Optional[int] = None
    profile: Optional[int] = None
    accountHolderName: Optional[str] = None
    type: Optional[str] = None
    country: Optional[str] = None
    currency: Optional[str] = None
    details: Optional[RecipientDetails] = None
    user: Optional[int] = None
    active: Optional[bool] = None
    ownedByCustomer: Optional[bool] = None
    endpoint: str = constants.ENDPOINT_RECIPIENT_ACCOUNTS_LIST

    @classmethod
    def call(cls, api_key: str, profile_id: int) -> List["Recipient"]:
        response: Response = http_requests.get(cls.endpoint.replace("{profile_id}", str(profile_id)), headers=get_header(api_key))
        return common.get_model_from_response(response, cls)  # type: ignore


@dataclass
class Fund(ResponseObject):
    type: Optional[str] = None
    status: Optional[str] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    balanceTransactionId: Optional[int] = None
    endpoint: str = constants.ENDPOINT_FUND

    @classmethod
    def call(cls, api_key: str, profile_id: int, transfer_id: int) -> "Fund":
        response: Response = http_requests.post(cls.endpoint.replace("{profile_id}", str(profile_id)).replace("{transfer_id}", str(transfer_id)), parameters={"type": "BALANCE"}, headers=get_header(api_key))
        return common.get_model_from_response(response, cls)  # type: ignore


@dataclass
class SourceAmount:
    value: Optional[float] = None
    currency: Optional[str] = None


@dataclass
class BalancesAfter:
    id: Optional[int] = None
    value: Optional[float] = None
    currency: Optional[str] = None
    cashAmount: Optional[SourceAmount] = None
    availableAmount: Optional[SourceAmount] = None
    totalWorth: Optional[SourceAmount] = None


@dataclass
class Step:
    tracingReferenceCode: None = None
    channelReferenceId: Optional[str] = None
    id: Optional[int] = None
    transactionId: Optional[int] = None
    type: Optional[str] = None
    creationTime: Optional[str] = None
    balancesAfter: Optional[List[SourceAmount]] = None
    channelName: Optional[str] = None
    sourceBalanceId: Optional[int] = None
    targetBalanceId: Optional[int] = None
    sourceAmount: Optional[SourceAmount] = None
    targetAmount: Optional[SourceAmount] = None
    fee: Optional[SourceAmount] = None
    rate: Optional[float] = None


@dataclass
class IntraAccountTransfer(ResponseObject):
    channelReferenceId: Optional[str] = None
    id: Optional[int] = None
    type: Optional[str] = None
    state: Optional[str] = None
    accountId: Optional[int] = None
    channelName: Optional[str] = None
    balancesAfter: Optional[List[BalancesAfter]] = None
    creationTime: Optional[str] = None
    steps: Optional[List[Step]] = None
    sourceAmount: Optional[SourceAmount] = None
    targetAmount: Optional[SourceAmount] = None
    rate: Optional[float] = None
    feeAmounts: Optional[List[SourceAmount]] = None
    endpoint: str = constants.ENDPOINT_INTRA_ACCOUNT_TRANSFER

    @classmethod
    def call(cls, api_key: str, profile_id: int, source_balance_id: int, target_balance_id: int, amount: float, quote_id: Optional[str], currency: str) -> "IntraAccountTransfer":
        parameters: Dict[str, Any] = {
            "profileId": profile_id,
            "sourceBalanceId": source_balance_id,
            "targetBalanceId": target_balance_id,
        }

        if quote_id is None:
            parameters["amount"] = {"value": amount, "currency": currency}
        else:
            parameters["quoteId"] = quote_id

        headers = get_header(api_key)
        headers["X-idempotence-uuid"] = str(uuid.uuid4())

        response: Response = http_requests.post(cls.endpoint.replace("{profile_id}", str(profile_id)), parameters=parameters, headers=headers)
        return common.get_model_from_response(response, cls)  # type: ignore


@dataclass
class AccountHolder:
    type: Optional[str] = None
    address: Optional[Address] = None
    firstName: Optional[str] = None
    lastName: Optional[str] = None


@dataclass
class EndOfStatementBalance:
    value: Optional[float] = None
    currency: Optional[str] = None
    zero: Optional[bool] = None


@dataclass
class Issuer:
    stateCode: None = None
    name: Optional[str] = None
    firstLine: Optional[str] = None
    city: Optional[str] = None
    postCode: Optional[str] = None
    countryCode: Optional[str] = None
    country: Optional[str] = None


@dataclass
class Query:
    intervalStart: Optional[str] = None
    intervalEnd: Optional[str] = None
    type: Optional[str] = None
    currency: Optional[str] = None
    profileId: Optional[int] = None
    timezone: Optional[str] = None


@dataclass
class Request:
    balanceName: None = None
    id: Optional[str] = None
    creationTime: Optional[str] = None
    profileId: Optional[int] = None
    currency: Optional[str] = None
    balanceId: Optional[int] = None
    intervalStart: Optional[str] = None
    intervalEnd: Optional[str] = None


@dataclass
class Details:
    type: Optional[str] = None
    description: Optional[str] = None
    sourceAmount: Optional[EndOfStatementBalance] = None
    targetAmount: Optional[EndOfStatementBalance] = None
    rate: Optional[float] = None
    recipient: Optional[Recipient] = None
    paymentReference: Optional[str] = None


@dataclass
class ExchangeDetails:
    toAmount: Optional[EndOfStatementBalance] = None
    fromAmount: Optional[EndOfStatementBalance] = None
    rate: Optional[float] = None


@dataclass
class Transaction:
    attachment: None = None
    type: Optional[str] = None
    date: Optional[str] = None
    amount: Optional[EndOfStatementBalance] = None
    totalFees: Optional[EndOfStatementBalance] = None
    details: Optional[Details] = None
    exchangeDetails: Optional[ExchangeDetails] = None
    runningBalance: Optional[EndOfStatementBalance] = None
    referenceNumber: Optional[str] = None
    activityAssetAttributions: Optional[List[Any]] = None

@dataclass
class AccountStatement(ResponseObject):
    endOfStatementUnrealisedGainLoss: None = None
    balanceAssetConfiguration: None = None
    accountHolder: Optional[AccountHolder] = None
    issuer: Optional[Issuer] = None
    bankDetails: Optional[List[Any]] = None
    transactions: Optional[List[Transaction]] = None
    endOfStatementBalance: Optional[EndOfStatementBalance] = None
    query: Optional[Query] = None
    request: Optional[Request] = None
    endpoint: str = constants.ENDPOINT_ACCOUNT_STATEMENT

    @classmethod
    def call(cls, api_key: str, profile_id: int, balance_id: int, start_time: datetime, end_time: datetime) -> "AccountStatement":
        parameters: Dict[str, str] = {
            "intervalStart": f"{start_time.isoformat()}Z",
            "intervalEnd": f"{end_time.isoformat()}Z"
        }
        response: Response = http_requests.get(cls.endpoint.replace("{profile_id}", str(profile_id)).replace("{balance_id}", str(balance_id)), parameters=parameters, headers=get_header(api_key))
        return common.get_model_from_response(response, cls)  # type: ignore


import os
from decimal import Decimal
from typing import Dict, Any, Optional

import pandas
from numpy import ndarray
from pandas import DataFrame

from kih_api.finance_database import common
from kih_api.finance_database.exceptions import DataNotFoundException


class ExcelData:
    excel_data: DataFrame
    data: ndarray
    rows_start_index: int
    rows_end_index: int
    columns_size: int
    row_size: int

    def __init__(self, excel_data: DataFrame):
        self.excel_data = excel_data
        self.data = self.excel_data.loc
        self.rows_start_index = self.excel_data.index.start
        self.rows_end_index = self.excel_data.index.stop
        self.columns_size = self.excel_data.columns.size
        self.row_size = self.rows_end_index - self.rows_start_index

    def get(self, row: int, columns: int) -> "Data":
        return Data(self.data[row][columns], row, columns)

    def find(self, query_string: str) -> "Data":
        for row in range(self.rows_start_index, self.rows_end_index):
            for column in range(0, self.columns_size):
                data: Data = self.get(row, column)
                if data.data == query_string:
                    return data
        raise DataNotFoundException()

    def get_data_set(self, query_string: str) -> Dict[str, Any]:
        header: Data = self.find(query_string)
        data_set: Dict[str, Any] = {}
        for row_number in range(1, max(self.columns_size, self.row_size)):
            try:
                key: str = self.get(row_number + header.row, header.column).data
                if key is not None:
                    data_set[key] = self.get(row_number + header.row, header.column + 1).data
                else:
                    raise KeyError
            except KeyError:
                return data_set
        return data_set

    def get_data_sub_set(self, query_string: str, number_of_datasets: int) -> "ExcelData":
        header: Data = self.find(query_string)
        end_column_number: int = header.column + (number_of_datasets * 2)
        end_row_number: Optional[int] = None

        row_number: int = header.row + 1
        while end_row_number is None:
            is_whole_row_nan: bool = True
            for column_number in range(header.column, end_column_number):
                is_whole_row_nan = is_whole_row_nan and self.get(row_number, column_number).data is None

            if is_whole_row_nan:
                end_row_number = row_number
            else:
                row_number = row_number + 1

        return ExcelData(self.excel_data.iloc[header.row:end_row_number, header.column:end_column_number])

    @staticmethod
    def read_excel(excel_file_path: str, sheet_name: str) -> "ExcelData":
        temp_file_path: str = common.get_temp_file_path(excel_file_path)
        excel_data: ExcelData = ExcelData(pandas.read_excel(temp_file_path, sheet_name))
        os.remove(temp_file_path)
        return excel_data


class Data:
    data: Any
    row: int
    column: int

    def __init__(self, data: Any, row: int, column: int):
        self.row = row
        self.column = column

        if pandas.isna(data):
            self.data = None
        elif isinstance(data, (int, float, complex)) and not isinstance(data, bool):
            self.data = Decimal(str(data))
        else:
            self.data = data


class Summary:
    header: str = "Summary"
    salary: Decimal
    savings: Decimal
    needs: Decimal
    wants: Decimal
    refuse: Decimal

    def __init__(self, excel_data: ExcelData):
        summary_data: Dict[str, Decimal] = excel_data.get_data_set(self.header)
        self.salary = summary_data["Salary"]
        self.savings = summary_data["Savings"]
        self.needs = summary_data["Needs"]
        self.wants = summary_data["Wants"]
        self.refuse = summary_data["Refuse"]
        self.assert_accuracy()

    def assert_accuracy(self) -> None:
        assert self.salary == (self.savings + self.needs + self.wants + self.refuse)


class MonthlyExpenseReport:
    header: str = "Monthly Expense Report"
    fixed_expenses: Decimal
    reserve: Decimal
    needs: Decimal
    wants: Decimal
    savings: Decimal
    extra_savings: Decimal

    def __init__(self, excel_data: ExcelData):
        monthly_expense_report_data: Dict[str, Decimal] = excel_data.get_data_set(self.header)
        self.fixed_expenses = monthly_expense_report_data["Total Fixed Expenses"]
        self.reserve = monthly_expense_report_data["Total Reserve"]
        self.needs = monthly_expense_report_data["Needs"]
        self.wants = monthly_expense_report_data["Wants"]
        self.savings = monthly_expense_report_data["Savings"]
        self.extra_savings = monthly_expense_report_data["Extra Savings"]

        self.assert_accuracy(excel_data)

    def assert_accuracy(self, excel_data: ExcelData) -> None:
        summary: Summary = Summary(excel_data)
        fixed_expenses: FixedExpenses = FixedExpenses(excel_data)
        reserve: Reserve = Reserve(excel_data)

        assert self.fixed_expenses == fixed_expenses.total
        assert self.reserve == reserve.total
        assert summary.salary == max(self.needs, Decimal("0")) + max(self.wants, Decimal("0")) + self.fixed_expenses + reserve.total + self.savings


class NeedsExpenses:
    header: str = "Needs"
    expenses: Dict[str, Decimal]
    total: Decimal

    def __init__(self, needs_expenses: Dict[str, Decimal]):
        self.expenses = needs_expenses

        self.total = Decimal("0")
        for expense in self.expenses.keys():
            self.total = self.total + self.expenses[expense]


class WantsExpenses:
    header: str = "Wants"
    expenses: Dict[str, Decimal]
    total: Decimal

    def __init__(self, wants_expenses: Dict[str, Decimal]):
        self.expenses = wants_expenses

        self.total = Decimal("0")
        for expense in self.expenses.keys():
            self.total = self.total + self.expenses[expense]


class FixedExpenses:
    header: str = "Fixed Expenses"
    needs_expenses: NeedsExpenses
    wants_expenses: WantsExpenses
    total: Decimal

    def __init__(self, excel_data: ExcelData):
        fixed_expenses_data: ExcelData = excel_data.get_data_sub_set(self.header, 2)
        self.needs_expenses = NeedsExpenses(fixed_expenses_data.get_data_set("Needs"))
        self.wants_expenses = WantsExpenses(fixed_expenses_data.get_data_set("Wants"))
        self.total = self.needs_expenses.total + self.wants_expenses.total


class Settings:
    settings: Dict[str, Any]

    def __init__(self, excel_data: ExcelData):
        self.settings = excel_data.get_data_set("Settings")

    def get(self, key: str) -> Any:
        return self.settings.get(key)


class Transfer:
    bank_account: str
    description: str
    amount: Decimal
    account_number: str

    def __init__(self, bank_account: str, description: str, amount: Decimal, account_number: str):
        self.bank_account = bank_account
        self.description = description
        self.amount = amount
        self.account_number = account_number.replace("-", "").strip()


class Transfers:
    finance_hub: Transfer
    needs: Transfer
    wants: Transfer
    savings: Transfer
    reserve: Transfer

    def __init__(self, excel_data: ExcelData, settings: Settings):
        data: Dict[str, Any] = excel_data.get_data_set("Transfers")

        item_number: int = 1
        for transfer in data.items():
            bank_account: str = transfer[0]
            amount: Decimal = transfer[1]
            if item_number == 1:
                self.finance_hub = Transfer(bank_account, "Finance Hub", amount, settings.get(bank_account + " Account Number"))
            elif item_number == 2:
                if amount > 0:
                    self.needs = Transfer(bank_account, "Needs", amount, settings.get(bank_account + " Account Number"))
                else:
                    self.needs = Transfer(bank_account, "Needs", Decimal(0), settings.get(bank_account + " Account Number"))
            elif item_number == 3:
                if amount > 0:
                    self.wants = Transfer(bank_account, "Wants", amount, settings.get(bank_account + " Account Number"))
                else:
                    self.wants = Transfer(bank_account, "Wants", Decimal(0), settings.get(bank_account + " Account Number"))
            elif item_number == 4:
                self.savings = Transfer(bank_account, "Savings", amount, settings.get(bank_account + " Account Number"))
            elif item_number == 5:
                self.reserve = Transfer(bank_account, "Reserve", amount, settings.get(bank_account + " Account Number"))
            item_number = item_number + 1

        assert Summary(
            excel_data).salary == self.finance_hub.amount + self.needs.amount + self.wants.amount + self.savings.amount + self.reserve.amount


class Reserve:
    header: str = "Reserve"
    needs_reserve: NeedsExpenses
    wants_reserve: WantsExpenses
    total: Decimal

    def __init__(self, excel_data: ExcelData):
        reserve_data: ExcelData = excel_data.get_data_sub_set(self.header, 2)
        self.needs_reserve = NeedsExpenses(reserve_data.get_data_set("Needs"))
        self.wants_reserve = WantsExpenses(reserve_data.get_data_set("Wants"))
        self.total = self.needs_reserve.total + self.wants_reserve.total



from kih_api.global_common import CustomException


class MultipleUserProfilesWithSameTypeException(CustomException):
    pass


class MultipleRecipientsWithSameAccountNumberException(CustomException):
    pass


class TransferringMoneyToNonSelfOwnedAccountsException(CustomException):
    pass


class ReserveAccountNotFoundException(CustomException):
    pass


class UnImplementedException(CustomException):
    pass



import enum
import inspect
import os
import subprocess
import threading
import time
from decimal import Decimal
from typing import Any, Dict, Type, Optional, List, Union, Callable

from dataclass_csv import DataclassWriter

from kih_api.logger import logger


class Currency(enum.Enum):
    USD: str = "USD"
    EUR: str = "EUR"
    GBP: str = "GBP"
    AUD: str = "AUD"
    NZD: str = "NZD"
    SGD: str = "SGD"
    LKR: str = "LKR"
    IDR: str = "IDR"

class Environment(enum.Enum):
    PROD: str = "Production"
    DEV: str = "Development"


class CustomException(Exception):
    pass


class EnumNotFoundException(CustomException):
    pass


def get_environment() -> Environment:
    try:
        if get_environment_variable("ENVIRONMENT") == Environment.PROD.value:
            return Environment.PROD
        else:
            return Environment.DEV
    except CustomException:
        return Environment.DEV

def get_environment_variable(key: str) -> str:
    value: str = os.getenv(key)
    if value is not None:
        return value
    else:
        raise CustomException(f"Environment variable not setup: {key}")

def convert_string_to_dict(string: str) -> Dict[Any, Any]:
    parameters_list = string.split(",")

    return_dict = {}
    for parameter in parameters_list:
        key_value_pair = parameter.split("=")
        key = key_value_pair[0]
        value = key_value_pair[1]

        return_dict[key] = value

    return return_dict


def get_enum_from_value(value: Any, enum: Type[enum.Enum]) -> Optional[Any]:
    for enum_value in enum:
        if enum_value.value == value:
            return enum_value

    raise EnumNotFoundException()


def get_formatted_string_from_decimal(number: Decimal, decimal_places: int = 2) -> str:
    return f'{number.quantize(Decimal(10) ** -decimal_places):,}'


def threaded(func: Callable) -> Callable:
    def wrapper(*args: Any, **kwargs: Any) -> threading.Thread:
        thread: threading.Thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        logger.debug(f"Thread started for function: {os.path.abspath(inspect.getfile(func))} | {func.__name__}")
        return thread

    return wrapper


def run_command(command_list: List[str]) -> Union[List[str], None]:
    output_list: List[str] = []

    for command in command_list:
        logger.info("Running command: " + command)
        output_list.extend(
            subprocess.run(command, shell=True, stdout=subprocess.PIPE).stdout.decode("utf-8").split("\n"))

    for output in output_list:
        logger.info(output)

    return output_list


def get_running_processes(command: str) -> List[Dict[str, str]]:
    results_list: List[str] = run_command([f"sudo pgrep -a {command}"])

    processes_list: List[Dict[str, str]] = []
    for result in results_list:
        if command in result:
            processes_list.append({"PID": result.split(" ")[0], "Command": result.replace(result.split(" ")[0], "")})

    return processes_list


def kill_process(keyword: str, command: str) -> bool:
    processes_list: List[Dict[str, str]] = get_running_processes(command)
    is_process_found_and_killed = False

    for process in processes_list:
        if keyword in process["Command"]:
            run_command([f"kill -9 {process['PID']}"])
            is_process_found_and_killed = True

    return is_process_found_and_killed


def update_code_base(project_directory: str, main_branch_name: str = "main") -> None:
    os.chdir(project_directory)

    run_command([f"git reset --hard origin/{main_branch_name}"])
    run_command([f"git pull origin {main_branch_name}"])

    if "requirements.txt" in os.listdir(project_directory):
        run_command([f"pip install -r requirements.txt"])


def job(job_name: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> None:
            from kih_api.communication import telegram
            try:
                telegram.send_message(
                    telegram.constants.telegram_debug_channel_username,
                    f"Running job: <i>{job_name}</i>", True)
                logger.debug(f"Running job: {job_name}")
                func(*args, **kwargs)
            except Exception as e:
                message = f"<b><u>ERROR</u></b>\n\nJob Name: <i>{job_name}</i>\nError Type: <i>{type(e).__name__}</i>"
                if str(e) != "":
                    message = message + f"\nError Message: <i>{str(e).replace('<', '').replace('>', '')}</i>"
                telegram.send_message(telegram.constants.telegram_channel_username, message,
                                                    True)
                raise Exception(e)
            finally:
                logger.debug(f"Job ended: {job_name}")
                telegram.send_message(
                    telegram.constants.telegram_debug_channel_username,
                    f"Job ended: <i>{job_name}</i>", True)

        return wrapper

    return decorator


def create_csv(data_class_list: List[Any], location: str) -> None:
    for data in data_class_list:
        if not isinstance(data, type(data_class_list[0])):
            raise Exception("Different objects types in data class list")

    with open(location, "w", newline="\n") as f:
        w = DataclassWriter(f, data_class_list, type(data_class_list[0]))
        w.write()


def timed(func: Callable) -> Callable:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        start_time: Decimal = Decimal(time.time_ns())
        return_data: Any = func(*args, **kwargs)
        end_time: Decimal = Decimal(time.time_ns())
        logger.debug(
            f"\n---------------------------------------------------------------------------------------------------"
            f"\nFunction: {func.__name__}"
            f"\nArguments: {args}"
            f"\nKey word arguments: {kwargs}"
            f"\nTime taken: {get_formatted_string_from_decimal(Decimal(end_time - start_time) / Decimal('1000_0000'))}ms"
            f"\n\t\t\t{get_formatted_string_from_decimal(Decimal(end_time - start_time) / Decimal('1000'))}ns"
            f"\n---------------------------------------------------------------------------------------------------")
        return return_data

    return wrapper



from __future__ import annotations

import threading
from decimal import Decimal
from typing import Dict, Any, Optional

threads: Dict[Any, threading.Thread] = {}


def extract_decimal(number: str | float) -> Optional[Decimal]:
    if isinstance(number, float):
        return Decimal(str(number))

    if number.replace(".", "").isnumeric():
        return Decimal(number)

    value: Decimal = Decimal(number.replace(number[-1], "").replace(",", ""))
    if number.upper().endswith("K"):
        return value * Decimal("1000")
    elif number.upper().endswith("M"):
        return value * Decimal("1000") * Decimal("1000")
    elif number.upper().endswith("B"):
        return value * Decimal("1000") * Decimal("1000") * Decimal("1000")
    elif number.upper().endswith("T"):
        return value * Decimal("1000") * Decimal("1000") * Decimal("1000")

    return None


def get_html_commented(error_html: str) -> str:
    if "<!DOCTYPE".lower() in error_html.lower():
        return "<!--" + error_html + "-->"
    else:
        return error_html



from kih_api.communication.telegram import constants
from kih_api.logger import logger

url = constants.telegram_url.replace("<token>", constants.telegram_bot_token) + constants.telegram_method_send_message


def send_message(username: str, message_text: str, is_html: bool = False) -> None:
    if username[0:1] != "@":
        username = "@" + username

    parameters = {"chat_id": username, "text": message_text}
    if is_html:
        parameters.update({"parse_mode": "HTML"})

    global url
    from kih_api import http_requests
    response = http_requests.post(url, parameters)

    if response.status_code >= 300:
        logger.error("Telegram message sending failed: " + response.json()["description"])



import logging as python_logging
import time

logger = python_logging.getLogger()
python_logging.basicConfig(level=python_logging.DEBUG, format="%(asctime)s | %(levelname)s | %(message)s")


def performance(execution_type: str, execution_subtype: str, execution_command: str, start_time: float, description: str = None) -> None:
    time_taken: int = int((time.time() - start_time) * 1000)
    log_message: str = str(time_taken) + "ms || Execution Type: " + execution_type + " | Execution Subtype:  " + execution_subtype

    if description is not None:
        log_message = log_message + " | " + description

    logger.debug(log_message)



from kih_api.global_common import CustomException


class KeepAliveFailedException(CustomException):
    pass


class AuthenticationFailedException(CustomException):
    pass


class StockNotFoundException(CustomException):
    pass


class MarketDataSnapshotNotAvailableException(CustomException):
    pass


class MarketOrderOutsideRTHException(CustomException):
    pass


class StockDataNotAvailableException(CustomException):
    pass



import datetime
import time

from kih_api import http_requests
from kih_api.ibkr_web_api import constants
from kih_api.ibkr_web_api.exceptions import AuthenticationFailedException
from kih_api.ibkr_web_api.ibkr_models import Authentication, KeepAlive, ReAuthentication, PortfolioAccounts, SelectAccount
from kih_api.logger import logger

http_requests.is_ssl_certificate_verification_used = False


def keep_alive() -> None:
    while True:
        logger.info("Keeping alive: " + str(datetime.datetime.now()))
        authentication: Authentication = Authentication.call()
        logger.info("Authentication status: " + str(authentication.authenticated))
        if not authentication.authenticated or not authentication.is_successful:
            if not authenticate():
                raise AuthenticationFailedException()
        time.sleep(constants.KEEP_ALIVE_TIMEOUT_IN_SECONDS)


def authenticate() -> bool:
    return ReAuthentication.call().is_successful




#   Telegram API
import os

from kih_api import global_common

telegram_url = "https://api.telegram.org/bot<token>/"
telegram_bot_token = global_common.get_environment_variable("KIH_API_TELEGRAM_BOT_TOKEN")
telegram_bot_username = "kontinuum_bot"
telegram_channel_username = global_common.get_environment_variable("TELEGRAM_CHANNEL_USERNAME") if global_common.get_environment() == global_common.Environment.PROD else global_common.get_environment_variable("TELEGRAM_CHANNEL_DEBUG_USERNAME")
telegram_debug_channel_username = global_common.get_environment_variable("TELEGRAM_CHANNEL_DEBUG_USERNAME")
telegram_stocks_data_template = "${Stock}: ${Stock-Price} | ${Stock-PriceChange}"
telegram_currencies_data_template = "${CurrencyPair}: ${ExchangeRate} | ${ExchangeRateChange}"

#   Telegram API - Methods
telegram_method_get_chat_id = "getChat"
telegram_method_send_message = "sendMessage"



import datetime
import enum
import time
from dataclasses import dataclass
from decimal import Decimal
from typing import Any, Dict, List, Union, Optional

import ibapi.common
import ibapi.contract
import ibapi.order
import ibapi.order_state

import kih_api
from kih_api import global_common, communication
from kih_api.communication import telegram
from kih_api.ibkr import constants, IBKR_API
from kih_api.ibkr.exceptions import MarketDataNotAvailableException
from kih_api.ibkr.helper import IBKR_Helper


class OrderType(enum.Enum):
    MARKET: str = "MKT"
    LIMIT: str = "LMT"


class OrderAction(enum.Enum):
    BUY: str = "BUY"
    SELL: str = "SELL"


class HistoricalDataType(enum.Enum):
    TRADES: str = "TRADES"
    MIDPOINT: str = "MIDPOINT"
    BID: str = "BID"
    ASK: str = "ASK"
    BID_ASK: str = "BID_ASK"
    ADJUSTED_LAST: str = "ADJUSTED_LAST"
    HISTORICAL_VOLATILITY: str = "HISTORICAL_VOLATILITY"
    OPTION_IMPLIED_VOLATILITY: str = "OPTION_IMPLIED_VOLATILITY"
    REBATE_RATE: str = "REBATE_RATE"
    FEE_RATE: str = "FEE_RATE"
    YIELD_BID: str = "YIELD_BID"
    YIELD_ASK: str = "YIELD_ASK"
    YIELD_BID_ASK: str = "YIELD_BID_ASK"
    YIELD_LAST: str = "YIELD_LAST"


class Exchange(enum.Enum):
    SMART: str = "SMART"
    NASDAQ: str = "NASDAQ"
    AMEX: str = "AMEX"
    NYSE: str = "NYSE"
    CBOE: str = "CBOE"
    NO_EXCHANGE: str = ""


class OrderStatus(enum.Enum):
    PENDING_SUBMIT: str = "PendingSubmit"
    PENDING_CANCEL: str = "PendingCancel"
    PRE_SUBMITTED: str = "PreSubmitted"
    SUBMITTED: str = "Submitted"
    API_CANCELLED: str = "ApiCancelled"
    CANCELLED: str = "Cancelled"
    FILLED: str = "Filled"
    INACTIVE: str = "Inactive"


class SecurityType(enum.Enum):
    STOCK: str = "STK"
    CFD: str = "CFD"
    INDEX: str = "IND"
    FUTURE: str = "FUT"
    OPTION: str = "OPT"
    FUTURES_OPTION: str = "FOP"
    BONDS: str = "BOND"
    MUTUAL_FUND: str = "FUND"
    COMMODITY: str = "CMDTY"
    STANDARD_WARRANT: str = "WAR"
    STRUCTURED_PRODUCT_AND_DUTCH_WARRANT: str = "IOPT"


class HistoricalDataBarSize(enum.Enum):
    SECOND_1: str = "1 secs"
    SECOND_5: str = "5 secs"
    SECOND_10: str = "10 secs"
    SECOND_15: str = "15 secs"
    SECOND_30: str = "30 secs"
    MINUTE_1: str = "1 min"
    MINUTE_2: str = "2 min"
    MINUTE_3: str = "3 min"
    MINUTE_5: str = "5 min"
    MINUTE_10: str = "10 min"
    MINUTE_15: str = "15 min"
    MINUTE_20: str = "20 min"
    MINUTE_30: str = "30 min"
    HOUR_1: str = "1 hour"
    HOUR_2: str = "2 hours"
    HOUR_3: str = "3 hours"
    HOUR_4: str = "4 hours"
    HOUR_8: str = "8 hours"
    DAY_1: str = "1 day"
    WEEK_1: str = "1 week"
    MONTH_1: str = "1 month"


class IBKR:
    @staticmethod
    def write_historical_data(symbol: str, location: str, bar_size: HistoricalDataBarSize = HistoricalDataBarSize.DAY_1, historical_data_type: HistoricalDataType = HistoricalDataType.TRADES) -> None:
        historical_data: List["HistoricalData"] = IBKR.get_historical_data(symbol, bar_size, historical_data_type)
        global_common.create_csv(historical_data, location)

    @staticmethod
    def get_historical_data(symbol: str, bar_size: HistoricalDataBarSize = HistoricalDataBarSize.DAY_1, historical_data_type: HistoricalDataType = HistoricalDataType.TRADES) -> List["HistoricalData"]:
        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.reqHistoricalData(ibkr_api.next_order_id, IBKR_Helper.get_contract_object(symbol, SecurityType.STOCK), (datetime.datetime.today()).strftime("%Y%m%d %H:%M:%S"), "30 Y", bar_size.value, historical_data_type.value, 1, 1, False, [])
        raw_historical_data_list: List[Dict[str, Any]] = IBKR_Helper.get_data_from_ibkr(ibkr_api, "historical_data_end", "historical_data")

        historical_data: List[HistoricalData] = []
        for raw_historical_data in raw_historical_data_list:
            historical_data.append(HistoricalData(raw_historical_data.get("bar")))
        return historical_data

    @staticmethod
    def is_market_open(symbol: str) -> bool:
        # try:
        #     IBKR.get_current_ask_price(symbol)
        #     return True
        # except MarketNotOpenException:
        #     return False
        return True

    @staticmethod
    def get_current_ask_price(symbol: str, number_of_tries: int = 0) -> Union[Decimal, None]:
        contract = IBKR_Helper.get_contract_object(symbol, SecurityType.STOCK)

        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.reqMktData(ibkr_api.next_order_id, contract, '', False, False, [])
        data_list: List[Dict[str, Any]] = IBKR_Helper.get_data_from_ibkr(ibkr_api, "ask_price")
        for data in data_list:
            if data.get("price") != -1:
                return Decimal(str(data.get("price")))

        if number_of_tries < 3:
            return IBKR.get_current_ask_price(symbol, number_of_tries + 1)
        else:
            raise MarketDataNotAvailableException()

    @staticmethod
    def place_order(symbol: str, quantity: Decimal, limit_price: Decimal, order_action: OrderAction, order_type: OrderType, security_type: SecurityType) -> "Order":
        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Placing a new order</b></u>\n\nSymbol: <i>{symbol}</i>"
                                            f"\nOrder Type: <i>{order_type.value}</i>"
                                            f"\nOrder Action: <i>{order_action.value}</i>"
                                            f"\nQuantity: <i>{str(quantity)}</i>"
                                            f"\nPrice: <i>{str(limit_price)}</i>", True)

        order: ibapi.order.Order = IBKR_Helper.get_order_object(quantity, limit_price, order_action, order_type)
        contract: ibapi.contract.Contract = IBKR_Helper.get_contract_object(symbol, security_type)

        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.placeOrder(ibkr_api.next_order_id, contract, order)

        data_list: List[Dict[str, Any]] = IBKR_Helper.get_data_from_ibkr(ibkr_api, "order_status")

        return Order(data_list[0]["orderId"], data_list[1]["status"], data_list[1]["filled"], data_list[1]["remaining"], data_list[1]["avgFillPrice"], data_list[0]["contract"], data_list[0]["order"], data_list[0]["orderState"])

    @staticmethod
    def close_all_positions() -> List["Order"]:
        IBKR.cancel_all_orders()

        positions_list: List[Position] = IBKR.get_positions()
        orders_list: List["Order"] = []
        for position in positions_list:
            if position.quantity > 0:
                telegram.send_message(telegram.constants.telegram_channel_username,
                                                    f"<u><b>Closing all positions</b></u>"
                                                    f"\n\nSymbol: <i>{position.symbol}</i>"
                                                    f"\nQuantity: <i>{str(position.quantity)}</i>", True)

                orders_list.append(IBKR.place_order(position.symbol, position.quantity, None, OrderAction.SELL, OrderType.MARKET, position.security_type))
        return orders_list

    @staticmethod
    def request_account_summary(currency: global_common.Currency = None) -> "Account":
        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()

        if currency is None:
            ibkr_api.reqAccountSummary(2, "All", "$LEDGER")
        else:
            ibkr_api.reqAccountSummary(2, "All", f"$LEDGER:{currency.value}")

        return Account(IBKR_Helper.get_data_from_ibkr(ibkr_api, "account_summary_end", "account_summary"))

    @staticmethod
    def get_positions() -> List["Position"]:
        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.reqPositions()

        data_list: List[Dict[str, Any]] = IBKR_Helper.get_data_from_ibkr(ibkr_api, "positions_end", "positions")
        position_list: List["Position"] = []

        if data_list is not None:
            for data in data_list:
                position_list.append(Position(data.get("account"), data.get("contract"), data.get("position"), data.get("avgCost")))

        return position_list

    @staticmethod
    def get_all_order_status() -> List[Any]:
        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.reqAllOpenOrders()

        return IBKR_Helper.get_data_from_ibkr(ibkr_api, "order_status_end", "order_status")

    @staticmethod
    def cancel_all_orders() -> None:
        telegram.send_message(telegram.constants.telegram_channel_username, "<u><b>Cancelling all orders</b></u>", True)

        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.reqGlobalCancel()
        time.sleep(constants.WAIT_TIME_IN_SECONDS)
        ibkr_api.disconnect()

    @staticmethod
    def get_all_managed_accounts() -> List[str]:
        ibkr_api: IBKR_API = IBKR_Helper.get_IBKR_connection()
        ibkr_api.reqManagedAccts()

        data_list: List[Any] = IBKR_Helper.get_data_from_ibkr(ibkr_api, "managed_accounts")
        return_data: List[str] = []
        for data in data_list:
            return_data.append(data.get("accountsList"))
        return return_data

    @staticmethod
    def is_connection_successful() -> bool:
        try:
            IBKR.get_current_ask_price("QQQ")
            return True
        except:
            return False


@dataclass
class HistoricalData:
    timestamp: datetime.datetime
    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    average: Decimal
    volume: Decimal

    def __init__(self, bar_data: ibapi.common.BarData):
        self.average = Decimal(str(bar_data.average))
        self.close = Decimal(str(bar_data.close))
        self.high = Decimal(str(bar_data.high))
        self.low = Decimal(str(bar_data.low))
        self.open = Decimal(str(bar_data.open))
        self.volume = Decimal(str(bar_data.volume))

        format_str: str = "%Y%m%d" if len(bar_data.date) == 8 else "%Y%m%d %H:%M:%S"
        self.timestamp = datetime.datetime.strptime(bar_data.date, format_str)

    @staticmethod
    def get_from_date(historical_data_list: List[HistoricalData], date: datetime.datetime, is_get_next_date_if_unavailable: bool = True) -> Optional[HistoricalData]:
        end_date_filter: datetime.datetime = date + datetime.timedelta(days=5 if is_get_next_date_if_unavailable else 0)
        filtered_historical_data_list: List[HistoricalData] = list(filter(lambda data: date <= data.timestamp <= end_date_filter, historical_data_list))
        sorted_historical_data_list: List[HistoricalData] = sorted(filtered_historical_data_list, key=lambda data: data.timestamp)

        if len(sorted_historical_data_list) != 0:
            return sorted_historical_data_list[0]
        else:
            return None


@dataclass
class Position:
    account: str
    symbol: str
    contract: "Contract"
    quantity: Decimal
    average_cost: Decimal
    security_type: SecurityType

    def __init__(self, account: str, contract: ibapi.contract.Contract, quantity: float, average_cost: float):
        self.account = account
        self.contract = Contract(contract.conId, contract.currency, contract.symbol, contract.exchange)
        self.symbol = self.contract.symbol
        self.quantity = Decimal(str(quantity))
        self.average_cost = Decimal(str(average_cost))
        self.security_type = global_common.get_enum_from_value(contract.secType, SecurityType)


@dataclass
class Contract:
    contract_id: int
    currency: global_common.Currency
    symbol: str
    exchange: Exchange

    def __init__(self, contract_id: int, currency: str, symbol: str, exchange: str):
        self.contract_id = contract_id
        self.currency = global_common.get_enum_from_value(currency, global_common.Currency)
        self.symbol = symbol
        self.exchange = global_common.get_enum_from_value(exchange, Exchange)


@dataclass
class Order:
    order_id: int
    status: OrderStatus
    quantity_filled: Decimal
    remaining_to_fill: Decimal
    average_fill_price: Decimal
    contract: Contract
    order: ibapi.order.Order
    order_state: ibapi.order_state.OrderState

    def __init__(self, order_id: int, status: str, quantity_filled: float, remaining_to_fill: float, average_fill_price: float, contract: ibapi.contract.Contract, order: ibapi.order.Order, order_state: ibapi.order_state.OrderState):
        self.order_id = order_id
        self.status = global_common.get_enum_from_value(status, OrderStatus)
        self.quantity_filled = Decimal(str(quantity_filled))
        self.remaining_to_fill = Decimal(str(remaining_to_fill))
        self.average_fill_price = Decimal(str(average_fill_price))
        self.contract = Contract(contract.conId, contract.currency, contract.symbol, contract.exchange)
        self.order = order
        self.order_state = order_state


@dataclass
class Account:
    cash_balance: Decimal
    stock_market_value: Decimal
    unrealized_pnl: Decimal
    net_liquidity: Decimal

    def __init__(self, data_list: List[Dict[str, Any]]):
        for data in data_list:
            if data.get("tag") == "TotalCashBalance":
                self.cash_balance = Decimal(str(data.get("value")))
            elif data.get("tag") == "StockMarketValue":
                self.stock_market_value = Decimal(str(data.get("value")))
            elif data.get("tag") == "UnrealizedPnL":
                self.unrealized_pnl = Decimal(str(data.get("value")))
            elif data.get("tag") == "NetLiquidationByCurrency":
                self.net_liquidity = Decimal(str(data.get("value")))



import os
import sys
from typing import Dict

from kih_api import global_common

environment_variables: Dict[str, str] = {"KIH_API_TELEGRAM_BOT_TOKEN": "telegram bot token",
                                         "TELEGRAM_BOT_USERNAME": "Telegram bot username",
                                         "KIH_API_EMAIL_ACCOUNT": "outgoing email address",
                                         "KIH_API_EMAIL_ACCOUNT_PASSWORD": "email address' password",
                                         "WISE_API_KEY": "Wise API Key",
                                         "ALPACA_API_KEY": "Alpaca API Key",
                                         "ALPACA_API_SECRET": "Alpaca API Secret",
                                         "MONGO_DB_URI": "The Mongo DB URL",
                                         "ENV": "The current environment"}


def set_environment_variable(key: str, value: str) -> None:
    if value == "":
        return
    elif sys.platform == "win32":
        global_common.run_command(["setx " + key + " \"" + value.replace("\"", "") + "\""])
    elif sys.platform == "linux":
        global_common.run_command([f"echo \"export {key}=\"{value}\" >> ~/.bash_profile"])
        global_common.run_command(["source ~/.bash_profile"])


if __name__ == "__main__":
    set_environment_variable("PYTHONPATH", os.getcwd())
    for key in environment_variables.keys():
        value: str = input("Enter the " + environment_variables[key] + ":")
        set_environment_variable(key, value)

print("\n\n----------------------------------------------------------------------------------")
print("Setup completed")
print("----------------------------------------------------------------------------------")



from dataclasses import dataclass
from typing import Optional

from requests import Response

import kih_api.http_requests.common
from kih_api import http_requests
from kih_api.global_common import constants
from kih_api.http_requests.models import ResponseObject


@dataclass
class IPAddressInfo(ResponseObject):
    zip: Optional[str] = None
    query: Optional[str] = None
    status: Optional[str] = None
    country: Optional[str] = None
    country_code: Optional[str] = None
    region: Optional[str] = None
    region_name: Optional[str] = None
    city: Optional[str] = None
    lat: Optional[float] = None
    lon: Optional[float] = None
    timezone: Optional[str] = None
    isp: Optional[str] = None
    org: Optional[str] = None
    google_maps_link: Optional[str] = None
    endpoint: str = constants.IP_ADDRESS_INFO_ENDPOINT

    @classmethod
    def get(cls, ip_address: str) -> "IPAddressInfo":
        response: Response = http_requests.get(cls.endpoint + ip_address)
        ip_address_info: IPAddressInfo = kih_api.http_requests.common.get_model_from_response(response, cls)  # type: ignore
        ip_address_info.google_maps_link = f"https://maps.google.com/?q={ip_address_info.lat},{ip_address_info.lon}"
        return ip_address_info



from kih_api.global_common import CustomException


class DataNotFoundException(CustomException):
    pass


class ExcelFileOpenedByAnotherApplication(CustomException):
    pass


class AccountForCurrencyNotFoundException(CustomException):
    pass


class InsufficientFundsException(CustomException):
    pass



import enum
from dataclasses import dataclass
from typing import Optional

from requests import Response


class MethodType(enum.Enum):
    GET: str = "GET"
    POST: str = "POST"
    PUT: str = "PUT"
    DELETE: str = "DELETE"


@dataclass
class ResponseObject:
    response: Optional[Response] = None
    is_successful: Optional[bool] = None
    endpoint: Optional[str] = None



import enum
import uuid
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import List, Union

import alpaca_trade_api.entity
import dateutil.parser
from alpaca_trade_api.rest import APIError

import kih_api
from kih_api import global_common, communication
from kih_api.alpaca import alpaca_api
from kih_api.communication import telegram


class AccountStatus(enum.Enum):
    ON_BOARDING: str = "ONBOARDING"
    SUBMISSION_FAILED: str = "SUBMISSION_FAILED"
    SUBMITTED: str = "SUBMITTED"
    ACCOUNT_UPDATED: str = "ACCOUNT_UPDATED"
    APPROVAL_PENDING: str = "APPROVAL_PENDING"
    ACTIVE: str = "ACTIVE"
    REJECTED: str = "REJECTED"
    DISABLE_PENDING: str = "DISABLE_PENDING"


class OrderSide(enum.Enum):
    BUY: str = "buy"
    SELL: str = "sell"


class OrderType(enum.Enum):
    LIMIT: str = "limit"
    MARKET: str = "market"
    STOP: str = "stop"
    STOP_LIMIT: str = "stop_limit"
    TRAILING_STOP: str = "trailing_stop"


class OrderTimeInForce(enum.Enum):
    DAY: str = "day"
    GTC: str = "gtc"
    OPG: str = "opg"
    CLS: str = "cls"
    IOC: str = "ioc"
    FOK: str = "fok"


class OrderStatus(enum.Enum):
    NEW: str = "new"
    PARTIALLY_FILLED: str = "partially_filled"
    FILLED: str = "filled"
    DONE_FOR_DAY: str = "done_for_day"
    CANCELED: str = "canceled"
    EXPIRED: str = "expired"
    REPLACED: str = "replaced"
    PENDING_CANCEL: str = "pending_cancel"
    ACCEPTED: str = "accepted"
    PENDING_NEW: str = "pending_new"
    ACCEPTED_FOR_BIDDING: str = "accepted_for_bidding"
    STOPPED: str = "stopped"
    REJECTED: str = "rejected"
    SUSPENDED: str = "suspended"
    CALCULATED: str = "calculated"


@dataclass
class Account:
    account_blocked: bool
    account_number: str
    buying_power: Decimal
    cash: Decimal
    currency: global_common.Currency
    daytrading_buying_power: Decimal
    equity: Decimal
    portfolio_value: Decimal
    status: AccountStatus

    def __init__(self, alpaca_account: alpaca_trade_api.entity.Account):
        self.account_blocked = alpaca_account._raw["account_blocked"]
        self.account_number = alpaca_account._raw["account_number"]
        self.buying_power = Decimal(str(alpaca_account._raw["buying_power"]))
        self.cash = Decimal(str(alpaca_account._raw["cash"]))
        self.currency = global_common.get_enum_from_value(alpaca_account._raw["currency"], global_common.Currency)
        self.daytrading_buying_power = Decimal(str(alpaca_account._raw["daytrading_buying_power"]))
        self.equity = Decimal(str(alpaca_account._raw["equity"]))
        self.portfolio_value = Decimal(str(alpaca_account._raw["portfolio_value"]))
        self.status = global_common.get_enum_from_value(alpaca_account._raw["status"], AccountStatus)

    @staticmethod
    def call() -> "Account":
        alpaca_account: alpaca_trade_api.entity.Account = alpaca_api.get_account()
        return Account(alpaca_account)


@dataclass
class Position:
    asset_id: str
    symbol: str
    exchange: str
    average_entry_price: Decimal
    quantity: Decimal
    market_value: Decimal
    cost_basis: Decimal
    current_price: Decimal
    change_in_percentage: Decimal
    unrealized_profit_loss: Decimal
    unrealized_profit_loss_in_percentage: Decimal

    def __init__(self, alpaca_position: alpaca_trade_api.entity.Position):
        self.asset_id = alpaca_position._raw["asset_id"]
        self.symbol = alpaca_position._raw["symbol"]
        self.exchange = alpaca_position._raw["exchange"]
        self.average_entry_price = Decimal(str(alpaca_position._raw["avg_entry_price"]))
        self.quantity = Decimal(str(alpaca_position._raw["qty"]))
        self.market_value = Decimal(str(alpaca_position._raw["market_value"]))
        self.cost_basis = Decimal(str(alpaca_position._raw["cost_basis"]))
        self.current_price = Decimal(str(alpaca_position._raw["current_price"]))
        self.change_in_percentage = Decimal(str(alpaca_position._raw["change_today"]))
        self.unrealized_profit_loss = Decimal(str(alpaca_position._raw["unrealized_pl"]))
        self.unrealized_profit_loss_in_percentage = Decimal(str(alpaca_position._raw["unrealized_plpc"]))

    @staticmethod
    def call() -> List["Position"]:
        alpaca_positions: List[alpaca_trade_api.entity.Position] = alpaca_api.list_positions()
        positions_list: List[Position] = []
        for alpaca_position in alpaca_positions:
            positions_list.append(Position(alpaca_position))
        return positions_list

    @staticmethod
    def get_by_symbol(symbol: str) -> "Position":
        try:
            alpaca_position: alpaca_trade_api.entity.Position = alpaca_api.get_position(symbol)
        except APIError as e:
            if str(e) == "position does not exist":
                return None
            else:
                raise APIError(e)

        return Position(alpaca_position)

    @staticmethod
    def close_all_by_symbol(symbol: str) -> Union["Order", None]:
        position: Position = Position.get_by_symbol(symbol)

        if position is not None:
            quantity: Decimal = position.quantity
        else:
            quantity = Decimal("0")

        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Closing whole position</b></u>"
                                            f"\n\nSymbol: <i>{symbol}</i>"
                                            f"\nQuantity: <i>{str(quantity)}</i>", True)

        try:
            alpaca_order: List[alpaca_trade_api.entity.Position] = alpaca_api.close_position(symbol)
        except APIError as e:
            if str(e) == "position does not exist":
                return None
            else:
                raise APIError(e)

        return Order(alpaca_order)

    @staticmethod
    def close_by_symbol(symbol: str, quantity: Decimal) -> List["Position"]:
        alpaca_positions: List[alpaca_trade_api.entity.Position] = alpaca_api.close_position(symbol, quantity)
        positions_list: List[Position] = []
        for alpaca_position in alpaca_positions:
            positions_list.append(Position(alpaca_position))
        return positions_list

    @staticmethod
    def close_all() -> List["Position"]:
        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Closing every single position</b></u>", True)

        alpaca_positions: List[alpaca_trade_api.entity.Position] = alpaca_api.close_all_positions()
        positions_list: List[Position] = []
        for alpaca_position in alpaca_positions:
            positions_list.append(Position(alpaca_position))
        return positions_list


@dataclass
class Order:
    id: str
    client_order_id: str
    created_at: datetime
    updated_at: datetime
    submitted_at: datetime
    filled_at: datetime
    expired_at: datetime
    canceled_at: datetime
    failed_at: datetime
    asset_id: str
    symbol: str
    quantity: Decimal
    filled_qty: Decimal
    filled_avg_price: Decimal
    order_type: OrderType
    side: OrderSide
    time_in_force: OrderTimeInForce
    limit_price: Decimal
    status: OrderStatus
    is_extended_hours: bool

    def __init__(self, alpaca_order: alpaca_trade_api.entity.Order):
        self.id = alpaca_order._raw["id"]
        self.created_at = dateutil.parser.isoparse(alpaca_order._raw["created_at"])
        self.updated_at = dateutil.parser.isoparse(alpaca_order._raw["updated_at"])
        self.submitted_at = dateutil.parser.isoparse(alpaca_order._raw["submitted_at"])
        self.asset_id = alpaca_order._raw["asset_id"]
        self.symbol = alpaca_order._raw["symbol"]
        self.quantity = Decimal(alpaca_order._raw["qty"])
        self.filled_qty = Decimal(alpaca_order._raw["filled_qty"])
        self.order_type = global_common.get_enum_from_value(alpaca_order._raw["order_type"], OrderType)
        self.side = global_common.get_enum_from_value(alpaca_order._raw["side"], OrderSide)
        self.time_in_force = global_common.get_enum_from_value(alpaca_order._raw["time_in_force"], OrderTimeInForce)
        self.status = global_common.get_enum_from_value(alpaca_order._raw["status"], OrderStatus)
        self.is_extended_hours = alpaca_order._raw["extended_hours"]

        if alpaca_order._raw["limit_price"] is not None:
            self.limit_price = Decimal(alpaca_order._raw["limit_price"])

        if alpaca_order._raw["filled_at"] is not None:
            self.filled_at = dateutil.parser.isoparse(alpaca_order._raw["filled_at"])

        if alpaca_order._raw["expired_at"] is not None:
            self.expired_at = dateutil.parser.isoparse(alpaca_order._raw["expired_at"])

        if alpaca_order._raw["canceled_at"] is not None:
            self.canceled_at = dateutil.parser.isoparse(alpaca_order._raw["canceled_at"])

        if alpaca_order._raw["failed_at"] is not None:
            self.failed_at = dateutil.parser.isoparse(alpaca_order._raw["failed_at"])

        if alpaca_order._raw["filled_avg_price"] is not None:
            self.filled_avg_price = Decimal(alpaca_order._raw["filled_avg_price"])

    @staticmethod
    def place(symbol: str, quantity: int, order_side: OrderSide, order_type: OrderType, limit_price: Decimal, time_in_force: OrderTimeInForce, custom_order_id: str = None) -> "Order":
        if custom_order_id is None:
            custom_order_id = str(uuid.uuid4())

        if order_type == OrderType.MARKET:
            float_limit_price: float = None
        else:
            float_limit_price = float(limit_price)

        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Placing a new order</b></u>"
                                            f"\n\nSymbol: <i>{symbol}</i>"
                                            f"\nOrder Type: <i>{order_type.value}</i>"
                                            f"\nOrder Side: <i>{order_side.value}</i>"
                                            f"\nQuantity: <i>{str(quantity)}</i>\nPrice: "
                                            f"<i>{str(float_limit_price)}</i>", True)

        alpaca_order: alpaca_trade_api.entity.Order = alpaca_api.submit_order(
            symbol=symbol,
            qty=int(quantity),
            side=order_side.value,
            type=order_type.value,
            time_in_force=time_in_force.value,
            limit_price=float_limit_price,
            client_order_id=custom_order_id
        )
        order: Order = Order(alpaca_order)

        if order.order_type == OrderType.MARKET:
            message: str = f"<u><b>Order status</b></u>" \
                           f"\n\nSymbol: <i>{symbol}</i>" \
                           f"\nOrder Type: <i>{order_type.value}</i>" \
                           f"\nOrder Side: <i>{order_side.value}</i>" \
                           f"\nQuantity: <i>{str(quantity)}</i>" \
                           f"\nFilled Quantity: <i>{str(order.filled_qty)}</i>"

            if order.filled_qty > 0:
                message = message + f"\nFilled Average Price: <i>${global_common.get_formatted_string_from_decimal(order.filled_avg_price)}</i>"

            telegram.send_message(telegram.constants.telegram_channel_username, message, True)

        return order

    @staticmethod
    def get_by_custom_order_id(custom_order_id: str) -> "Order":
        alpaca_order = alpaca_api.get_order_by_client_order_id(custom_order_id)
        return Order(alpaca_order)

    @staticmethod
    def get_all() -> List["Order"]:
        alpaca_order_list: List[alpaca_trade_api.entity.Order] = alpaca_api.list_orders()
        order_list: List["Order"] = []
        for alpaca_order in alpaca_order_list:
            order_list.append(Order(alpaca_order))
        return order_list

    @staticmethod
    def cancel_by_custom_order_id(custom_order_id: str) -> None:
        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Cancelling Order</b></u>"
                                            f"\n\nCustom Order ID: <i>{custom_order_id}</i>", True)

        alpaca_api.cancel_order(custom_order_id)

    @staticmethod
    def cancel_all() -> None:
        telegram.send_message(telegram.constants.telegram_channel_username,
                                            f"<u><b>Cancelling all orders</b></u>", True)

        alpaca_api.cancel_all_orders()


@dataclass
class MarketDataHistory:
    timestamp: datetime
    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    volume: Decimal

    def __init__(self, alpaca_market_data: alpaca_trade_api.entity.Bar):
        self.timestamp = datetime.utcfromtimestamp(alpaca_market_data._raw["t"])
        self.open = Decimal(str(alpaca_market_data._raw["o"]))
        self.high = Decimal(str(alpaca_market_data._raw["h"]))
        self.low = Decimal(str(alpaca_market_data._raw["l"]))
        self.close = Decimal(str(alpaca_market_data._raw["c"]))
        self.volume = Decimal(str(alpaca_market_data._raw["v"]))

    @staticmethod
    def get_history(symbol: str, number_of_days: int = None) -> List["MarketDataHistory"]:
        alpaca_market_data_history_list: alpaca_trade_api.entity.Bars = alpaca_api.get_barset(symbol, "day", limit=number_of_days)[symbol]
        market_data_history_list: List[MarketDataHistory] = []
        for alpaca_market_data in alpaca_market_data_history_list:
            market_data_history_list.append(MarketDataHistory(alpaca_market_data))
        return market_data_history_list


@dataclass
class MarketData:
    symbol: str
    ask_price: Decimal
    ask_size: Decimal
    bid_price: Decimal
    bid_size: Decimal
    timestamp: datetime

    def __init__(self, symbol: str, alpaca_market_data: alpaca_trade_api.entity.Quote):
        self.symbol = symbol
        self.ask_price = Decimal(str(alpaca_market_data._raw["askprice"]))
        self.ask_size = Decimal(str(alpaca_market_data._raw["asksize"]))
        self.bid_price = Decimal(str(alpaca_market_data._raw["bidprice"]))
        self.bid_size = Decimal(str(alpaca_market_data._raw["bidsize"]))
        self.timestamp = datetime.utcfromtimestamp(int(alpaca_market_data._raw["timestamp"] / 1000000000))

    @staticmethod
    def get(symbol: str) -> "MarketData":
        alpaca_market_data: alpaca_trade_api.entity.Quote = alpaca_api.get_last_quote(symbol)
        return MarketData(symbol, alpaca_market_data)


@dataclass
class MarketClock:
    timestamp: datetime
    is_open: bool
    next_open: datetime
    next_close: datetime

    def __init__(self, alpaca_clock: alpaca_trade_api.entity.Clock):
        self.timestamp = dateutil.parser.isoparse(alpaca_clock._raw["timestamp"])
        self.is_open = alpaca_clock._raw["is_open"]
        self.next_open = dateutil.parser.isoparse(alpaca_clock._raw["next_open"])
        self.next_close = dateutil.parser.isoparse(alpaca_clock._raw["next_close"])

    @staticmethod
    def get() -> "MarketClock":
        alpaca_clock: alpaca_trade_api.entity.Clock = alpaca_api.get_clock()
        return MarketClock(alpaca_clock)



from datetime import datetime

from kih_api.finance_database import common
from kih_api.finance_database.models import ExcelData, Transfers, FixedExpenses, MonthlyExpenseReport, Summary, Settings, Reserve


class FinanceDatabase:
    settings: Settings
    summary: Summary
    monthly_expenses_report: MonthlyExpenseReport
    fixed_expenses: FixedExpenses
    transfers: Transfers
    reserve: Reserve

    def __init__(self, excel_file_path: str, month: datetime = None):
        excel_data: ExcelData = ExcelData.read_excel(excel_file_path, common.get_next_month_text() if month is None else month.strftime("%B, %Y"))
        self.settings = Settings(excel_data)
        self.summary = Summary(excel_data)
        self.monthly_expenses_report = MonthlyExpenseReport(excel_data)
        self.fixed_expenses = FixedExpenses(excel_data)
        self.transfers = Transfers(excel_data, self.settings)
        self.reserve = Reserve(excel_data)



import os
from typing import Dict

from kih_api import global_common

ENDPOINT_BASE_SANDBOX: str = "https://api.transferwise.com/" if global_common.get_environment() == global_common.Environment.PROD else "https://api.sandbox.transferwise.tech/"
ENDPOINT_BASE_LIVE: str = "https://api.transferwise.com/"
ENDPOINT_BASE: str = ENDPOINT_BASE_SANDBOX
ENDPOINT_PROFILES: str = ENDPOINT_BASE + "v1/profiles"
ENDPOINT_ACCOUNTS: str = ENDPOINT_BASE + "v4/profiles/{profile_id}/balances?types={account_type}"
ENDPOINT_EXCHANGE_RATES: str = ENDPOINT_BASE + "v1/rates"
ENDPOINT_TRANSFER: str = ENDPOINT_BASE + "v1/transfers"
ENDPOINT_QUOTE: str = ENDPOINT_BASE + "v3/profiles/{profile_id}/quotes"
ENDPOINT_RECIPIENT_ACCOUNTS_LIST: str = ENDPOINT_BASE + "v1/accounts?profile={profile_id}"
ENDPOINT_FUND: str = ENDPOINT_BASE + "v3/profiles/{profile_id}/transfers/{transfer_id}/payments"
ENDPOINT_CREATE_ACCOUNT: str = ENDPOINT_BASE + "v3/profiles/{profile_id}/balances"
ENDPOINT_INTRA_ACCOUNT_TRANSFER: str = ENDPOINT_BASE + "v2/profiles/{profile_id}/balance-movements"
ENDPOINT_ACCOUNT_STATEMENT: str = ENDPOINT_BASE + "v1/profiles/{profile_id}/balance-statements/{balance_id}/statement.json"


from __future__ import annotations

import enum
import typing
from abc import ABC
from dataclasses import dataclass
from datetime import datetime, timedelta
from decimal import Decimal
from functools import cache
from typing import List, Optional, Union

import pytz

from kih_api import global_common
from kih_api.communication import telegram
from kih_api.finance_database.exceptions import InsufficientFundsException, AccountForCurrencyNotFoundException
from kih_api.http_requests import ClientErrorException
from kih_api.logger import logger
from kih_api.wise import wise_models
from kih_api.wise.exceptions import MultipleUserProfilesWithSameTypeException, \
    MultipleRecipientsWithSameAccountNumberException, TransferringMoneyToNonSelfOwnedAccountsException, \
    ReserveAccountNotFoundException


class ProfileType(enum.Enum):
    Personal: str = "personal"
    Business: str = "business"


class AccountType(enum.Enum):
    CashAccount: str = "STANDARD"
    ReserveAccount: str = "SAVINGS"


@dataclass
class WiseAccount:
    api_key: str
    profile_type: ProfileType
    _user_profile: Optional[UserProfile] = None
    _account_list: Optional [List[CashAccount | ReserveAccount]] = None
    _recipient_list: Optional[List[Recipient]] = None

    def __init__(self, api_key_environmental_variable_key: str, profile_type: ProfileType):
        self.api_key = global_common.get_environment_variable(api_key_environmental_variable_key)
        self.profile_type = profile_type

    @property
    def user_profile(self) -> UserProfile:
        if self._user_profile is not None:
            return self._user_profile

        self._user_profile = UserProfile.get_by_profile_type(self.api_key, self.profile_type)
        return self._user_profile

    @property
    def account_list(self) -> List[CashAccount | ReserveAccount]:
        if self._account_list is not None:
            return self._account_list

        self._account_list = Account.get_all_by_profile_type(self.api_key, self.profile_type)
        return self._account_list


    def get_cash_account(self, currency: global_common.Currency) -> CashAccount:
        return typing.cast(CashAccount, list(filter(lambda account: account.currency == currency and isinstance(account, CashAccount), self.account_list))[0])

    def get_reserve_account(self, currency: global_common.Currency, reserve_account_name: str, create_if_unavailable: bool = False) -> ReserveAccount:
        reserve_account: ReserveAccount = typing.cast(ReserveAccount, list(filter(lambda account: account.currency == currency and isinstance(account, ReserveAccount) and account.name == reserve_account_name,self.account_list))[0])

        if reserve_account is None:
            return ReserveAccount.create_reserve_account(self.api_key, reserve_account_name, currency, self.profile_type)

        return reserve_account

    def get_exchange_rate(self, from_currency: global_common.Currency, to_currency: global_common.Currency) -> ExchangeRate:
        return ExchangeRate.get(self.api_key, from_currency, to_currency)

    @property
    def recipient_list(self) -> List[Recipient]:
        if self._recipient_list is not None:
            return self._recipient_list

        self._recipient_list = Recipient.get_all_by_profile_type(self.api_key, self.profile_type)
        return self._recipient_list

    def get_recipient_by_account_number(self, account_number: str) -> Recipient:
        return list(filter(lambda recipient: recipient.account_number == account_number , self.recipient_list))[0]

    def get_all_transactions(self, account: Account, start_time: datetime, end_time: datetime) -> List[Transaction]:
        return Transaction.get_all(self.api_key, account, start_time, end_time)

@dataclass
class UserProfile:
    id: int
    type: ProfileType
    first_name: str
    last_name: str
    date_of_birth: str

    @classmethod
    def get_all(cls, api_key: str) -> List["UserProfile"]:
        user_profiles_list: List[UserProfile] = []

        for wise_user_profile in wise_models.UserProfiles.call(api_key):
            user_profiles_list.append(UserProfile(wise_user_profile.id, global_common.get_enum_from_value(wise_user_profile.type, ProfileType), wise_user_profile.details.firstName, wise_user_profile.details.lastName, wise_user_profile.details.dateOfBirth))

        return user_profiles_list

    @classmethod
    def get_by_profile_type(cls, api_key: str, profile_type: ProfileType) -> "UserProfile":
        return_user_profile: UserProfile = None

        for user_profile in UserProfile.get_all(api_key):
            if user_profile.type == profile_type:
                if return_user_profile is None:
                    return_user_profile = user_profile
                else:
                    raise MultipleUserProfilesWithSameTypeException()
        return return_user_profile


@dataclass
class Account(ABC):
    id: int
    currency: global_common.Currency
    balance: Decimal
    type: AccountType
    name: str
    account_type: AccountType
    user_profile: UserProfile
    api_key: str

    def __init__(self, api_key: str, wise_account: wise_models.Account, user_profile: UserProfile):
        self.api_key = api_key
        self.id = wise_account.id
        self.currency = global_common.get_enum_from_value(wise_account.currency, global_common.Currency)
        self.balance = Decimal(str(wise_account.cashAmount.value))
        self.type = global_common.get_enum_from_value(wise_account.type, AccountType)
        self.account_type = global_common.get_enum_from_value(wise_account.type, AccountType)
        self.user_profile = user_profile

    def transfer(self, recipient: Recipient, receiving_amount: Decimal, reference: str = "", to_currency: Optional[global_common.Currency] = None) -> Transfer:
        to_currency = self.currency if to_currency is None else to_currency
        transfer: Transfer = Transfer.execute(self.api_key, receiving_amount, recipient, self.currency, to_currency, reference, self.user_profile.type)

        if transfer.is_successful:
            if isinstance(self, CashAccount):
                self.balance = Account.get_by_profile_type_and_currency(self.api_key, self.user_profile.type, self.currency).balance
            else:
                self.balance = ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(self.api_key, self.user_profile.type, self.currency, self.name, False).balance

        return transfer

    def intra_account_transfer(self, to_account: Union[CashAccount | ReserveAccount], receiving_amount: Decimal) -> IntraAccountTransfer:
        intra_account_transfer: IntraAccountTransfer = IntraAccountTransfer.execute(self.api_key, receiving_amount, self, to_account, self.user_profile.type)    # type: ignore[arg-type]
        if intra_account_transfer.is_successful:
            if isinstance(self, CashAccount):
                self.balance = Account.get_by_profile_type_and_currency(self.api_key, self.user_profile.type, self.currency).balance
            else:
                self.balance = ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(self.api_key, self.user_profile.type, self.currency, self.name, False).balance

            if isinstance(to_account, CashAccount):
                to_account.balance = Account.get_by_profile_type_and_currency(self.api_key, self.user_profile.type, self.currency).balance
            else:
                to_account.balance = ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(self.api_key, self.user_profile.type, self.currency, self.name).balance

        return intra_account_transfer

    @classmethod
    def get_all_by_profile_type(cls, api_key: str, profile_type: ProfileType) -> List["CashAccount" | "ReserveAccount"]:
        user_profile: UserProfile = UserProfile.get_by_profile_type(api_key, profile_type)
        accounts_list: List["CashAccount" | "ReserveAccount"] = []

        for wise_account in wise_models.Account.call(api_key, user_profile.id):
            account_type: AccountType = global_common.get_enum_from_value(wise_account.type, AccountType)
            if account_type == AccountType.CashAccount:
                accounts_list.append(CashAccount(api_key, wise_account, user_profile))
            elif account_type == AccountType.ReserveAccount:
                accounts_list.append(ReserveAccount(api_key, wise_account, user_profile))

        return accounts_list

    @classmethod
    def get_by_profile_type_and_currency(cls, api_key: str, profile_type: ProfileType, currency: global_common.Currency) -> "CashAccount" | "ReserveAccount":
        all_accounts_list: List["CashAccount" | "ReserveAccount"] = Account.get_all_by_profile_type(api_key, profile_type)
        for account in all_accounts_list:
            if account.currency == currency:
                return account
        raise AccountForCurrencyNotFoundException()


@dataclass
class ReserveAccount(Account):
    name: str

    def __init__(self, api_key: str, wise_account: wise_models.Account, user_profile: UserProfile):
        super().__init__(api_key, wise_account, user_profile)
        self.name = wise_account.name

    @classmethod
    def get_all_by_profile_type(cls, api_key: str, profile_type: ProfileType) -> List["ReserveAccount"]:  # type: ignore
        all_accounts: List["CashAccount" | "ReserveAccount"] = super().get_all_by_profile_type(api_key, profile_type)
        all_cash_accounts: List[ReserveAccount] = []

        for account in all_accounts:
            if isinstance(account, ReserveAccount):
                all_cash_accounts.append(account)
        return all_cash_accounts

    @classmethod
    def get_by_profile_type_and_currency(cls, api_key: str, profile_type: ProfileType, currency: global_common.Currency) -> List["ReserveAccount"]:  # type: ignore
        all_accounts_list: List[ReserveAccount] = ReserveAccount.get_all_by_profile_type(api_key, profile_type)
        reserve_account_list: List[ReserveAccount] = []
        for account in all_accounts_list:
            if account.currency == currency:
                reserve_account_list.append(account)
        return reserve_account_list

    @classmethod
    @cache
    def get_by_cached_profile_type_and_currency(cls, api_key: str, profile_type: ProfileType, currency: global_common.Currency) -> List["ReserveAccount"]:
        return ReserveAccount.get_by_profile_type_and_currency(api_key, profile_type, currency)

    @classmethod
    def get_reserve_account_by_profile_type_currency_and_name(cls, api_key: str, profile_type: ProfileType, currency: global_common.Currency, name: str, create_if_unavailable: bool = False) -> "ReserveAccount":
        all_reserve_accounts_list: List[ReserveAccount] = ReserveAccount.get_by_profile_type_and_currency(api_key, profile_type, currency)
        for reserve_account in all_reserve_accounts_list:
            if reserve_account.name == name:
                return reserve_account
        if create_if_unavailable:
            return ReserveAccount.create_reserve_account(api_key, name, currency, profile_type, False)
        else:
            raise ReserveAccountNotFoundException()

    @classmethod
    def create_reserve_account(cls, api_key: str, name: str, currency: global_common.Currency, profile_type: ProfileType, check_if_available_before_creation: bool = True) -> "ReserveAccount":
        if check_if_available_before_creation:
            try:
                return ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(api_key, profile_type, currency, name)
            except ReserveAccountNotFoundException:
                wise_models.Account.call_create_reserve_account(api_key, name, currency, UserProfile.get_by_profile_type(api_key, profile_type).id)

        return ReserveAccount.get_reserve_account_by_profile_type_currency_and_name(api_key, profile_type, currency, name)

@dataclass
class CashAccount(Account):

    def __init__(self, api_key: str, wise_account: wise_models.Account, user_profile: UserProfile):
        super().__init__(api_key, wise_account, user_profile)

    @classmethod
    def get_all_by_profile_type(cls, api_key: str, profile_type: ProfileType) -> List["CashAccount"]:  # type: ignore
        all_accounts: List["CashAccount" | "ReserveAccount"] = super().get_all_by_profile_type(api_key, profile_type)
        all_cash_accounts: List[CashAccount] = []

        for account in all_accounts:
            if isinstance(account, CashAccount):
                all_cash_accounts.append(account)
        return all_cash_accounts

    @classmethod
    def get_by_profile_type_and_currency(cls, api_key: str, profile_type: ProfileType, currency: global_common.Currency) -> "CashAccount":
        all_accounts_list: List[CashAccount] = CashAccount.get_all_by_profile_type(api_key, profile_type)
        for account in all_accounts_list:
            if account.currency == currency:
                return account
        raise AccountForCurrencyNotFoundException()


@dataclass
class ExchangeRate:
    exchange_rate: Decimal
    from_currency: global_common.Currency
    to_currency: global_common.Currency

    @classmethod
    def get(cls, api_key: str, from_currency: global_common.Currency, to_currency: global_common.Currency) -> "ExchangeRate":
        wise_exchange_rate: wise_models.ExchangeRate = wise_models.ExchangeRate.call(api_key, from_currency.value, to_currency.value)
        return ExchangeRate(Decimal(str(wise_exchange_rate.rate)), from_currency, to_currency)


@dataclass
class Recipient:
    account_id: int
    profile_id: int
    name: str
    currency: global_common.Currency
    is_active: bool
    is_self_owned: bool
    account_number: str
    swift_code: str
    bank_name: str
    branch_name: str
    iban: str
    bic: str

    @classmethod
    def get_all_by_profile_type(cls, api_key: str, profile_type: ProfileType) -> List["Recipient"]:
        user_profile: UserProfile = UserProfile.get_by_profile_type(api_key, profile_type)
        recipient_list: List[Recipient] = []

        for wise_recipient in wise_models.Recipient.call(api_key, user_profile.id):
            recipient_list.append(Recipient(wise_recipient.id, wise_recipient.profile, wise_recipient.accountHolderName, global_common.get_enum_from_value(wise_recipient.currency, global_common.Currency), wise_recipient.active, wise_recipient.ownedByCustomer, wise_recipient.details.accountNumber, wise_recipient.details.swiftCode, wise_recipient.details.bankName, wise_recipient.details.branchName, wise_recipient.details.swiftCode, wise_recipient.details.bic))

        return recipient_list

    @classmethod
    def get_by_account_number_and_profile_type(cls, api_key: str, account_number: str, profile_type: ProfileType) -> "Recipient":
        return_recipient: Recipient = None
        for recipient in Recipient.get_all_by_profile_type(api_key, profile_type):
            if recipient.account_number == account_number:
                if return_recipient is None:
                    return_recipient = recipient
                else:
                    raise MultipleRecipientsWithSameAccountNumberException()

        return return_recipient


@dataclass
class Transfer:
    profile_id: int
    from_currency: global_common.Currency
    to_currency: global_common.Currency
    from_amount: Decimal
    to_amount: Decimal
    recipient: Recipient
    exchange_rate: ExchangeRate
    is_successful: bool
    error_message: Optional[str]
    error_code: Optional[str]

    def __init__(self, user_profile: UserProfile, recipient: Recipient, wise_transfer: wise_models.Transfer, wise_fund: wise_models.Fund):
        self.profile_id = user_profile.id
        self.from_currency = global_common.get_enum_from_value(wise_transfer.sourceCurrency, global_common.Currency)
        self.to_currency = global_common.get_enum_from_value(wise_transfer.targetCurrency, global_common.Currency)
        self.from_amount = Decimal(str(wise_transfer.sourceValue))
        self.to_amount = Decimal(str(wise_transfer.targetValue))
        self.recipient = recipient
        self.exchange_rate = ExchangeRate(Decimal(str(wise_transfer.rate)), self.from_currency, self.to_currency)
        self.is_successful = wise_fund.status == "COMPLETED"
        self.error_message = wise_fund.errorMessage
        self.error_code = wise_fund.errorCode

    @classmethod
    def execute(cls, api_key: str, receiving_amount: Decimal, recipient: Recipient, from_currency: global_common.Currency, to_currency: global_common.Currency, reference: str, profile_type: ProfileType) -> "Transfer":
        if not recipient.is_self_owned:
            raise TransferringMoneyToNonSelfOwnedAccountsException()

        try:
            if from_currency == to_currency:
                cash_account: CashAccount = CashAccount.get_by_profile_type_and_currency(api_key, profile_type, from_currency)
                if cash_account.balance < receiving_amount:
                    raise InsufficientFundsException(f"Insufficient funds"
                                                     f"\nRequired amount: {from_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}"
                                                     f"\nAccount Balance: {from_currency.value} {global_common.get_formatted_string_from_decimal(cash_account.balance)}"
                                                     f"\nShort of: {from_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount - cash_account.balance)}")

            user_profile: UserProfile = UserProfile.get_by_profile_type(api_key, profile_type)
            wise_quote: wise_models.Quote = wise_models.Quote.call(api_key, user_profile.id, from_currency.value, to_currency.value, float(receiving_amount))
            wise_transfer: wise_models.Transfer = wise_models.Transfer.call(api_key, recipient.account_id, wise_quote.id, reference)
            wise_fund: wise_models.Fund = wise_models.Fund.call(api_key, user_profile.id, wise_transfer.id)
            transfer: Transfer = Transfer(user_profile, recipient, wise_transfer, wise_fund)

            if transfer.is_successful:
                telegram.send_message(telegram.constants.telegram_channel_username,
                                                    f"<u><b>Money transferred</b></u>"
                                                    f"\n\nAmount: <i>{to_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                    f"\nTo: <i>{recipient.name}</i>"
                                                    f"\nReference: <i>{reference}</i>", True)
            else:
                raise ClientErrorException(transfer.error_message)

            return transfer

        except ClientErrorException as e:
            logger.error(str(e))
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nTo: <i>{recipient.name}</i>"
                                                f"\nReason: <i>{str(e)}</i>"
                                                f"\nReference: <i>{reference}</i>", True)
            raise e
        except InsufficientFundsException as e:
            logger.error(str(e))
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nTo: <i>{recipient.name}</i>"
                                                f"\nReference: <i>{reference}</i>"
                                                f"\n\nReason: <i>{str(e)}</i>", True)
            raise e


@dataclass
class IntraAccountTransfer:
    from_currency: global_common.Currency
    from_amount: Decimal
    from_account: Account
    to_currency: global_common.Currency
    to_amount: Decimal
    to_account: Account
    exchange_rate: ExchangeRate
    is_successful: bool

    def __init__(self, from_account: Account, to_account: Account, intra_account_transfer: wise_models.IntraAccountTransfer):
        self.from_currency = global_common.get_enum_from_value(intra_account_transfer.sourceAmount.currency, global_common.Currency)
        self.from_amount = Decimal(str(intra_account_transfer.sourceAmount.value))
        self.from_account = from_account
        self.to_account = to_account
        self.to_currency = global_common.get_enum_from_value(intra_account_transfer.targetAmount.currency, global_common.Currency)
        self.to_amount = Decimal(str(intra_account_transfer.targetAmount.value))
        self.exchange_rate = ExchangeRate(Decimal(str(intra_account_transfer.rate)), self.from_currency, self.to_currency)
        self.is_successful = intra_account_transfer.state == "COMPLETED"

    @classmethod
    def execute(cls, api_key: str, receiving_amount: Decimal, from_account: Union[CashAccount, ReserveAccount], to_account: Union[CashAccount, ReserveAccount], profile_type: ProfileType) -> "IntraAccountTransfer":
        user_profile: UserProfile = UserProfile.get_by_profile_type(api_key, profile_type)

        if receiving_amount < Decimal("0"):
            to_account, from_account = from_account, to_account
            receiving_amount = receiving_amount * Decimal("-1")

        try:
            if isinstance(to_account, CashAccount):
                return IntraAccountTransfer._transfer_to_cash_account(api_key, receiving_amount, from_account, to_account, user_profile)
            else:
                return IntraAccountTransfer._transfer_to_reserve_account(api_key, receiving_amount, typing.cast(CashAccount, from_account), to_account, user_profile)

        except ClientErrorException as e:
            logger.error(str(e))
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Intra account money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nFrom: <i>{to_account.currency.value}</i>"
                                                f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else None} ({to_account.currency.value})</i>"
                                                f"\nReason: <i>{str(e)}</i>", True)
            raise e
        except InsufficientFundsException as e:
            logger.error(str(e))
            telegram.send_message(telegram.constants.telegram_channel_username,
                                                f"<u><b>ERROR: Intra account money transfer failed</b></u>"
                                                f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                                                f"\nFrom: <i>{to_account.currency.value}</i>"
                                                f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else to_account.currency.value} ({to_account.currency.value})</i>"
                                                f"\nReason: <i>{str(e)}</i>", True)
            raise e

    @classmethod
    def _transfer_to_cash_account(cls, api_key: str, receiving_amount: Decimal, from_account: Union[CashAccount, ReserveAccount], to_account: CashAccount, user_profile: UserProfile) -> "IntraAccountTransfer":
        if isinstance(from_account, ReserveAccount):
            wise_intra_account_transfer = wise_models.IntraAccountTransfer.call(api_key, user_profile.id, from_account.id, to_account.id, float(receiving_amount), None, to_account.currency.value)
        else:
            wise_quote: wise_models.Quote = wise_models.Quote.call(api_key, user_profile.id, from_account.currency.value,to_account.currency.value, float(receiving_amount))
            wise_intra_account_transfer = wise_models.IntraAccountTransfer.call(api_key, user_profile.id, from_account.id, to_account.id, float(receiving_amount), wise_quote.id if isinstance(from_account, CashAccount) else None, None)

        intra_account_transfer = IntraAccountTransfer(from_account, to_account, wise_intra_account_transfer)

        if intra_account_transfer.is_successful:
            telegram.send_message(
                telegram.constants.telegram_channel_username,
                f"<u><b>Intra account money transferred</b></u>"
                f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                f"\nFrom: <i>{from_account.name if isinstance(from_account, ReserveAccount) else to_account.currency.value}</i>"
                f"\nTo: <i>{to_account.currency.value}</i>", True)

        return intra_account_transfer

    @classmethod
    def _transfer_to_reserve_account(cls, api_key: str, receiving_amount: Decimal, from_account: CashAccount, to_account: ReserveAccount, user_profile: UserProfile) -> "IntraAccountTransfer":
        if from_account.currency != to_account.currency:
            to_cash_account: CashAccount = CashAccount.get_by_profile_type_and_currency(api_key, user_profile.type, to_account.currency)
            IntraAccountTransfer.execute(api_key, receiving_amount, from_account, to_cash_account, user_profile.type)
            from_account = to_cash_account

        if from_account.balance < receiving_amount:
            raise InsufficientFundsException(f"Insufficient funds"
                                             f"\nRequired amount: {from_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}"
                                             f"\nAccount Balance: {from_account.currency.value} {global_common.get_formatted_string_from_decimal(from_account.balance)}"
                                             f"\nShort of: {from_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount - from_account.balance)}")

        wise_intra_account_transfer = wise_models.IntraAccountTransfer.call(api_key, user_profile.id, from_account.id, to_account.id, float(receiving_amount), None, to_account.currency.value)
        intra_account_transfer = IntraAccountTransfer(from_account, to_account, wise_intra_account_transfer)

        if intra_account_transfer.is_successful:
            telegram.send_message(
                telegram.constants.telegram_channel_username,
                f"<u><b>Intra account money transferred</b></u>"
                f"\n\nAmount: <i>{to_account.currency.value} {global_common.get_formatted_string_from_decimal(receiving_amount)}</i>"
                f"\nFrom: <i>{to_account.currency.value}</i>"
                f"\nTo: <i>{to_account.name if isinstance(to_account, ReserveAccount) else ''} ({to_account.currency.value})</i>", True)

        return intra_account_transfer


class TransactionType(enum.Enum):
    Card: str = "CARD"
    Balance: str = "BALANCE"
    Transfer: str = "TRANSFER"
    WISE_AUTOMATED_TRANSACTION: str = "OPS"

@dataclass
class Transaction:
    transaction_type: TransactionType
    timestamp: datetime
    currency: global_common.Currency
    total_amount: Decimal
    fees: Decimal
    transaction_amount: Decimal
    running_balance: Decimal
    reference: Optional[str]
    entity: Optional[str | Account]
    profile_type: ProfileType

    def __init__(self, wise_transaction: wise_models.Transaction, profile_type: ProfileType):
        self.profile_type = profile_type
        self.transaction_type = global_common.get_enum_from_value(wise_transaction.referenceNumber.split("-")[0] , TransactionType)
        self.timestamp = pytz.utc.localize(datetime.fromisoformat(wise_transaction.date.split(".")[0]))
        self.currency = global_common.get_enum_from_value(wise_transaction.runningBalance.currency, global_common.Currency)
        self.total_amount = Decimal(str(wise_transaction.amount.value))
        self.fees = Decimal(str(wise_transaction.totalFees.value))
        self.transaction_amount = self.total_amount - self.fees if self.total_amount > Decimal("0") else self.total_amount + self.fees
        self.running_balance = Decimal(str(wise_transaction.runningBalance.value))
        self.reference = wise_transaction.details.paymentReference
        self.get_entity(wise_transaction)

    def get_entity(self, wise_transaction: wise_models.Transaction) -> None:
        if self.transaction_type == TransactionType.Transfer:
            self.entity = wise_transaction.details.description.replace("Received money from ", "").replace(" with reference ", "").replace("Sent money to ","")
        elif self.transaction_type == TransactionType.Card:
            self.entity = wise_transaction.details.description.split("issued by ")[1]
        elif self.transaction_type == TransactionType.Balance:
            reserve_account_name: str = wise_transaction.details.description.split(" to ")[1] if "to" in wise_transaction.details.description else wise_transaction.details.description.split(" from ")[1]
            try:
                self.entity = list(filter(lambda reserve_account: reserve_account.name == reserve_account_name, ReserveAccount.get_by_cached_profile_type_and_currency(self.profile_type, self.currency)))[0]
            except IndexError:
                self.entity = None



    @classmethod
    def get_all(cls, api_key: str, account: Account, start_time: datetime = datetime.now() - timedelta(days=1), end_time: datetime = datetime.now()) -> List["Transaction"]:
        wise_account_statement: wise_models.AccountStatement = wise_models.AccountStatement.call(api_key, account.user_profile.id, account.id, start_time, end_time)
        transaction_list: List[Transaction] = []

        for transaction in wise_account_statement.transactions:
            transaction_list.append(Transaction(transaction, account.user_profile.type))

        return transaction_list


import datetime
import time
import typing
from decimal import Decimal
from typing import List, Any, Union

from ibapi.contract import Contract
from ibapi.order import Order

from kih_api.ibkr import IBKR_API, connect_to_ib_api
from kih_api.ibkr.exceptions import IBKR_APITimeOutException

if typing.TYPE_CHECKING:
    from kih_api.ibkr.models import OrderAction, OrderType, HistoricalData, SecurityType


class IBKR_Helper:
    @staticmethod
    def get_data_from_ibkr(ibkr_api: IBKR_API, key_to_wait_for: str, data_key: str = None) -> List[Any]:
        data: List = None
        for _ in range(1, 60):
            if ibkr_api.data.get(key_to_wait_for) is not None:
                data = ibkr_api.data.get(key_to_wait_for) if data_key is None else ibkr_api.data.get(data_key)
                break
            else:
                time.sleep(1)

        ibkr_api.disconnect()

        if ibkr_api.data.get(key_to_wait_for) is None:
            raise IBKR_APITimeOutException()

        return data

    @staticmethod
    def get_IBKR_connection() -> IBKR_API:
        ibkr_api = IBKR_API()
        connect_to_ib_api(ibkr_api)
        return ibkr_api

    @staticmethod
    def get_contract_object(symbol: str, security_type: "SecurityType") -> Contract:
        contract = Contract()
        contract.symbol = symbol
        contract.secType = security_type.value
        contract.exchange = "SMART"
        contract.currency = "USD"

        from kih_api.ibkr.models import SecurityType
        if security_type == SecurityType.STOCK:
            contract.primaryExchange = "NASDAQ"
        return contract

    @staticmethod
    def get_order_object(quantity: Decimal, limit_price: Decimal, order_action: "OrderAction", order_type: "OrderType") -> Order:
        order = Order()
        order.action = order_action.value
        order.orderType = order_type.value
        order.totalQuantity = int(quantity)
        order.outsideRth = True

        if order_type.value != "MKT" or limit_price is not None:
            order.lmtPrice = float(limit_price)
        return order


class HistoricalDataHelper:
    @staticmethod
    def filter_historical_data_list(starting_date: Union[datetime.datetime, str], ending_date: Union[datetime.datetime, str], historical_data_list: List["HistoricalData"]) -> List["HistoricalData"]:

        if isinstance(starting_date, str):
            starting_date = datetime.datetime.strptime(starting_date, '%Y-%m-%d')

        if ending_date is None:
            ending_date = datetime.datetime.today()
        elif isinstance(ending_date, str):
            ending_date = datetime.datetime.strptime(ending_date, '%Y-%m-%d')

        return_historical_data: List["HistoricalData"] = []
        for historical_data in historical_data_list:
            if starting_date <= historical_data.timestamp <= ending_date:
                return_historical_data.append(historical_data)

        return_historical_data.sort(key=lambda x: x.timestamp)
        return return_historical_data



